/*****************************************************************************
* CE v1.2 Copyright (C) 1991-92, Charles R. Henrich, All Rights Reserved.    *
*                                                                            *
* Use in source and binary forms, with or without modification and           *
* Redistribution without modification are permitted provided that the        *
* following conditions are met:                                              *
*                                                                            *
* 1. Redistributions of source code must retain the above copyright          *
*    notice, all internal copyright notices, this list of conditions         *
*    and the following disclaimer.                                           *
* 2. Redistributions in binary form must reproduce the above copyright       *
*    notice, this list of conditions and the following disclaimer in the     *
*    documentation and/or other materials provided with the distribution.    *
* 3. All advertising materials mentioning features or use of this software   *
*    must display the following acknowledgement:                             *
*      CE Was developed in 1991-92 by Charles Henrich                        *
* 4. Neither the name of this product (CE) nor the name of the author        *
*    (Charles Henrich) may be used to endorse products that include this     *
*    product without specific prior written permission.                      *
* 5. No profit may be made from the inclusion, or distribution of this       *
*    product.  Although a reasonable service charge for the distribution     *
*    of this product is permitted where appropriate.                         *
*                                                                            *
* THIS SOFTWARE IS PROVIDED BY CHARLES HENRICH `` AS IS '' AND ANY EXPRESS   *
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED          *
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE     *
* DISCLAIMED.  IN NO EVENT SHALL CHARLES HENRICH BE LIABLE FOR ANY DIRECT,   *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES         *
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR         *
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)         *
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT *
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY  *
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF     *
* SUCH DAMAGE.                                                               *
*                                                                            *
******************************************************************************
*                                                                            *
* This file contains all of the functions used by CE.                        *
*                                                                            *
*****************************************************************************/

#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "ce.h"
#ifndef NO_STDLIB
#include <stdlib.h>
#endif

extern int MARKING;
extern int LINES;
extern int COLUMNS;
extern int MAXLINE;
extern int FILELEN;
extern int STARTBUF;
extern int ENDBUF;
extern char STATUSLINE[];
extern char STATUSMSG[];
extern char STATUSMODE[];
extern int wordwrap;
extern int tabsize;
extern int G_mini;
extern int G_noscroll;

char *itoa(int var);
void clean_exit(int return_code);
int wrapline(int locx, int locy, int vlocy, char *lines[], int lcol, int *ret);
char *getstr(int locx, int locy, char *string);
void readrc(int *insertmode,int *tabsize,int *lcol,
     char *SHELL1, char *SHELL2, int *wordwrap, int *nobaks, int *G_mini,
     int *G_noscroll);
int gotoline(int linenum, int *locy, int *locx, int *vlocy, char *lines[]);
void savefile(char *fname, int wordwrap, int quit, int nobaks);
#ifdef NORINDEX
char *index(char *s, char c);
char *rindex(char *s, char c);
#endif
#ifdef NOSTRSTR
char *strstr(char *s1, char *s2);
#endif
void drawstatus(int locx, int locy);
void remap(int *cha, int *func);
void scrolldown(int *locx, int *locy, int *vlocy, char *lines[]);
void scrollup(int *locx, int *locy, int *vlocy, char *lines[]);
void search(char string[],int *locy,int *locx, int *vlocy);
void copy(char fname[90], char nname[90]);
void getseq(int *p, int ret);
void keyset(int killit);
void keyfilename(char tname[]);
void RedrawLine(int locx, int locy, int vlocy);
void getdata(char msg[], char def[], char tmp[],
     int locx, int locy, int vlocy, char *lines[], int strip);
void RedrawFromX(int locx, int locy, char *line);
int insertchars(int num, int locx, char ch, char *line);
int deletechars(int num, int locx, char *line);
int insertlines(char *lines[], int loc, int num);
int deletelines(char *lines[], int loc, int num);
void strip(char *lines[], int line, int wordwrap);
void RedrawScreen(int top, char *lines[], int locx, int locy);
void padlines(int vlocy, char *lines[]);
void padchars(int start, int end, char lines[]);
int seqcmp(int q[], int seqtable[NUMELM]);
void clrq(int q[]);
void keyinit();
int arsize(int q[]);
void shiftseq(int q[], int n);
void gettimeout(int sig);
int getkeyortimeout();

jmp_buf get_time_buf;

int NUMSEQ = 0;
int seqtable[MAXNUMSEQ][NUMELM];

/* This function is the raw, low-level keyboard handling access routine
   all keyboard access is retrieved through here.  This function will get
   a single character and compare it to all entries in the special-key table
   (seqtable) if a possible match is found, it then asks for another character
   until eaither a) it finds a complete match, if it does it sets the value
   in func to any special key listed in the table (and referenced in the header
   file) or b) if it doesnt find a match, it returns all the characters to the
   main program */

void getkey(int *cha, int *func)
{

/* q is the input que, it is cleared whenever a valid match is found,
   or when the program has determined that a match cannot happen */

static int q[NUMELM] = {0};
static initialize = 1;
register int count, count1;
int qlen, possible;

/* Clear the input que only on program startup */

if(initialize)
    {
    clrq(q);
    initialize=0;
    }

/* With no data a match is always possible */

possible=1;

/* Do until a match is impossible, or found */

while(1)
    {
    qlen = arsize(q);

    /* Is this the first character typed?
       If so, is it a normal Character? (I.e.
       A-Z a-z 0-9 CRLF, or TAB... If it is
       return the type of key it is */

    if (qlen > 0)
        {
        if ((q[0] > 31 && q[0] < 127))
            {
            *cha=q[0];
            *func=PRINTABLE;
            shiftseq(q,1);
            return;
            }

        /* There is no possible match yet, we start the
            checking here */

        possible=0;

        /* Search the first element of the special-key table
           if a match, go on to compare all sub-elements, until
           eaither a match is proved impossible, or until we need
           more data */

        for(count=0;count<NUMSEQ;count++)
            {
            if (q[0] == seqtable[count][1])
                {
                possible=seqcmp(q,seqtable[count]);
                if (possible == MATCH)
                    {
                    *func=seqtable[count][0];
                    clrq(q);
                    return;
                    }
                }
            if (possible)
                {
                break;
                }
            }
        if (!possible)
            {
            break;
            }
        }

    /* If the input que is empty get a character
       or if a possible sequence is matched get the next
       character, if there isnt one in one-second timeout */

    if (qlen == 0)
        {
        q[qlen] = getchar();
        }
    else
        {
        if(possible)
            {
            q[qlen] = getkeyortimeout();

            if (q[qlen] == TIMEOUT)
                {
                q[qlen]=0;
                break;
                }
            }
        }
    }

/* A match is NOT possible start to send back the characters, one at
   at time...  (is the first character in the que "special" ?? */

if (q[0] < 27 && q[0] > 0)
    {
    *cha=q[0]+64;
    *func=CONTROL;
    shiftseq(q,1);
    return;
    }

if (q[0] == 27)
    {
    *func=ESC;
    if(q[1] != 0) *cha = MOREDATA;
    shiftseq(q,1);
    return;
    }

/* If it makes it this far, it isnt anything we know, so make it unknown */

*func=UNKNOWN;
*cha=-1;
shiftseq(q,1);
return;

}

/* This function compares two integer arrays, and returns
   MATCH if they are identicle, or 1 if more data is needed
   or 0 if no match possible */

int seqcmp(int q[], int seqtable[NUMELM])
{
register int count;

for(count=1;count < NUMELM; count++)
    {
    if (q[count-1] == seqtable[count]) continue;
    break;
    }

if (count == NUMELM)
    {
    return MATCH;
    }

if (q[count-1] == 0)
    {
    return 1;
    }
return 0;
}

/* This function clears the input que to 0 */

void clrq(int q[])
{
register int count;

for(count=0;count < NUMELM; count++)
    {
    q[count]=0;
    }
}

/* This function sets up the special-key table */

void keyinit()
{
register int count, count1;
char home[100];
char fnam[20];
char fname[120];
int ret;
FILE *ifp;

signal(SIGALRM, gettimeout);

/* Get the environment variable for this terminal's name */

if (strncpy(home, getenv("HOME"),100) == NULL) strcpy(home,".");
keyfilename(fnam);

strcat(home,"/.ce");
strcpy(fname,home);
strcat(fname,"/");
strcat(fname,fnam);

/* Try and read the special-key table in the users home dir */

ifp=fopen(fname,"r");

/* If not here look in the default path for the file */

if(ifp == NULL)
    {
    strcpy(fname,CONFIG_PATH);
    strcat(fname,fnam);
    ifp=fopen(fname,"r");

    if (ifp == NULL)
        {
#ifdef FORCE_KEYSET
        keyset(0);
        strcpy(fname,home);
        strcat(fname,"/");
        strcat(fname,fnam);
        ifp=fopen(fname,"r");

        if(ifp == NULL)
            {
            ClearScreen();
         printf("\aSomething is REALLY messed up, go get your sys admin!\r\n");
            clean_exit(-1);
            }
#else
        ClearScreen();
        printf("\n\r");
        printf("Unable to locate a key map for your terminal.\n\r");
        printf("To create one please run ce with the command\n\r");
        printf("\n\r");
        printf("   ce -keyset\n\r");
        printf("\n\r");
        printf("You may also want to contact your system administrator\n\r");
        printf("To install a keymap for your terminal for system use\n\r");
        printf("\r\n\a");
        clean_exit(0);
#endif
        }
    }

/* Read in the data */

for(count=0;count<MAXNUMSEQ;count++)
    {
    for(count1=0;count1<NUMELM;count1++)
        {
        ret=fscanf(ifp,"%d",&seqtable[count][count1]);
        }
    if(ret != 1) break;
    }
NUMSEQ=count;
fclose(ifp);
}

/* This function finds ths size of a one dimension array */

int arsize(int q[])
{
register int count;
for(count=0;count<NUMELM;count++)
    {
    if (q[count] == 0)
        {
        break;
        }
    }
return count;
}

/* This function copies all elements of an array n positions to the left
   and fills the remaining spaces with 0 */

void shiftseq(int q[], int n)
{
register int count;

for (count=0;count < (NUMELM-n); count++)
    {
    q[count]=q[count+n];
    }

for(count=(NUMELM-n);count>NUMELM;count++)
    {
    q[count]=0;
    }
}

/* Timeout code, re-initialize after trigger */

void gettimeout(int sig)
{
signal(SIGALRM, gettimeout);
longjmp(get_time_buf, 1);
}

/* Get a key, or return TIMEOUT if nothing gotten in one second */

int getkeyortimeout()
{
int rtn;
if (setjmp(get_time_buf)!=0)
    {
    alarm(0);
    return TIMEOUT;
    }
alarm(1);
rtn = getchar();
alarm(0);
return rtn;
}

/* Ask for 80 bytes of memory to store one more line of text in...
   Possible *bad* thing is that by allocating memory this way will
   fragment memory *really* bad after awhile (too bad! :) ) */

void padlines(int vlocy, char *lines[])
{
register int count;
char *cptr;

if(vlocy > MAXLINE) return;

if(lines[vlocy] != 0) return;

for (count=FILELEN;count<=vlocy;count++)
    {
    if ( lines[count] != NULL ) continue;
    cptr = (char *)calloc(CALLOCSIZE, sizeof(char));
    if ( cptr == NULL )
        {
        fprintf(stderr,"Out of memory in padlines - ce aborted\n");
        majerr();
        }
    lines[count] = cptr;
    }
FILELEN=vlocy;
}

/* Fill a character array with n blanks characters, starting from pos 0 */

void padchars(int start, int end, char lines[])
{
register int count;
for(count=start;count<end;count++)
    {
    lines[count]=' ';
    }
lines[count]='\0';
}

/* This function will clear the screen and repaint it with data in the
   array passed to it...  Also check for screen size changes... */

void RedrawScreen(int top, char *lines[], int locx, int locy)
{
int toggle;
int yloc;
register int count;
int bottom;

if(top<0)top=0;
yloc = 0;
toggle=0;

ClearScreen();

drawstatus(0,0);

/* Make sure we dont try and paint from a non-existant part of the array */

if(FILELEN < (LINES+top))
    {
    bottom=FILELEN;
    }
else
    {
    bottom=LINES+top;
    }

/* Do the actual painting */

for(count=top;count<=bottom;count++)
    {
    toggle = 0;
    MoveCursor(0,yloc);
    if(count >= STARTBUF && count <= ENDBUF) StartBold();
    if(strlen(lines[count]) > (COLUMNS-2))
        {
        if(lines[count][COLUMNS - 1] == 32 && wordwrap == 0)
            {
            lines[count][COLUMNS - 1] = '>';
            toggle=1;
            }
        }
    printf("%s",lines[count]);
    if(strlen(lines[count]) > (COLUMNS-2))
        if(toggle) lines[count][COLUMNS-1]=32;
    if(count >= STARTBUF && count <= ENDBUF) EndBold();
    yloc++;
    }
MoveCursor(locx,locy);
}

/* Take out all nasty characters from the array.  "Nasty" characters
   are ones this editor doenst like, and they are all control characters
   as well as tab's and other such characters.  Only plain alphanumeric
   characters will be tolerated */

void strip(char *lines[], int line, int wordwrap)
{
register int count;
int ret;
int flag;
flag=0;
for(count = strlen(lines[line]);count >= 0; count--)
    {
    if (isgraph(lines[line][count]))
        {
        flag=1;
        }
    if (lines[line][count] == '\n')
        {
        lines[line][count] = '\0';
        }
    if (lines[line][count] == '\t')
        {
        lines[line][count]=' ';
        ret=insertchars(tabsize-1,count,' ',lines[line]);
        }
    if(wordwrap != 0 && lines[line][count] == ' ' && !flag)
        {
        lines[line][count] = '\0';
        }
    if (lines[line][count] <32 && lines[line][count] >0)
        lines[line][count]=' ';
    }
}

/* Allocate memory for n lines, and put them at x location in the pointer
   array */

int insertlines(char *lines[], int loc, int num)
{
register int count, count1;

if ( ( (num+FILELEN) > MAXLINE)  || (loc > MAXLINE) )
    {
    return -1;
    }

for(count=MAXLINE;count>=loc+num;count--)
    {
    lines[count]=lines[count-num];
    }

for(count=loc;count<loc+num;count++)
    {
    lines[count]=(char *)calloc(CALLOCSIZE, sizeof(char));
    }

FILELEN+=num;

if(STARTBUF != -1 && ENDBUF != -1)
    {
    if(loc <= STARTBUF)
        {
        STARTBUF+=num;
        if(STARTBUF > FILELEN) STARTBUF = FILELEN;
        ENDBUF+=num;
        if(ENDBUF > FILELEN) ENDBUF = FILELEN;
        }
    else
        {
        if(loc > STARTBUF && loc <= ENDBUF)
            {
            ENDBUF+=num;
            if(ENDBUF > FILELEN) ENDBUF = FILELEN;
            }
        }
    }
return 1;
}

/* Wipe out n lines from the pointer array at x location, and also free
   up the unused memory.  *BAD* fragmentation will be caused here too
   but such is life... :) */

int deletelines(char *lines[], int loc, int num)
{
register int count, count1;

if (loc > FILELEN) return -1;
if ((loc+num-1) > FILELEN) return -1;
if (num == 0) return -1;

for(count=loc;count<=MAXLINE-num;count++)
    {
    if(lines[count] != 0)
        {
        free(lines[count]);
        }
    lines[count]=lines[count+num];
    lines[count+num]=0;
    }

FILELEN-=num;

if(STARTBUF != -1 && ENDBUF != -1)
    {
    if(loc < STARTBUF)
        {
        STARTBUF-=num;
        if(STARTBUF < 0) STARTBUF=0;
        ENDBUF-=num;
        if(ENDBUF < 0) ENDBUF=0;
        }
    else
        {
        if(loc >= STARTBUF && loc <= ENDBUF)
            {
            ENDBUF-=num;
            if(ENDBUF < 0) ENDBUF=0;

            }
        }
    }
return 1;
}

/* Insert n characters ch at x location */

int insertchars(int num, int locx, char ch, char *line)
{
register int count;

if(line == NULL) return -1;

if(locx < 0) return -1;

if(strlen(line) + num > COLUMNS) return -1;

if(num < 0) return -1;

for(count=COLUMNS-1;count>=locx;count--)
    {
    line[count]=line[count-num];
    }

for(count=locx;count<locx+num;count++)
    {
    line[count]=ch;
    }

return 1;
}

/* Redraw part of the current line, from location x to end */

void RedrawFromX(int locx, int locy, char *line)
{
register int count;

if(line == NULL) return;

for(count=locx; line[count]; count++)
    {
    putchar(line[count]);
    }
}

/* Delete x characters at this location (from the array) */

int deletechars(int num, int locx, char *line)
{
register int count, count1;

if(line == NULL) return -1;

if(locx+num > strlen(line)) num=strlen(line)-locx;

for(count=0; count<num;count++)
    {
    for(count1=locx;line[count1];count1++)
        {
        line[count1]=line[count1+1];
        }
    line[count1]='\0';
    }
return 1;
}

/* This function gets data from the user in a "pretty" manner */

void getdata(char *msg, char *def, char *tmp,
     int locx, int locy, int vlocy, char *lines[], int strip)
{
int count;
int toggle;
int itmp;
char ctmp[90];

toggle =0;
ctmp[0] = tmp[0] = '\0';

/* Erase the top two lines of the screen */

MoveCursor(0,0);
CleartoEOLN();
MoveCursor(0,1);
CleartoEOLN();
MoveCursor(0,0);

/* Print the request string & and the default answer */

printf("%s:%s",msg,def);

/* Move the cursor the the beginning of the default answer */

MoveCursor(strlen(msg)+1,0);

if(getstr(strlen(msg)+1,0,ctmp) == NULL)
    {
    tmp[0]='\0';
    }

/* Strip "bad" data from the array the user entered, this time only straight
   alphanumeric, no special characters */

if(strip)
    {
    for(count=0;ctmp[count];count++)
        {
        if((ctmp[count] < 46) ||
           (ctmp[count] > 57 && ctmp[count] < 65) ||
           (ctmp[count] > 90 && ctmp[count] < 97) ||
           (ctmp[count] > 122 && ctmp[count] != 126))
            {
            ctmp[count]='\0';
            }
        }
    }
else
    {
    for(count=0;ctmp[count];count++)
        {
        if(ctmp[count] < 32) ctmp[count]='\0';
        }
    }

if(ctmp[0]=='~')
    {
    strncpy(tmp, getenv("HOME"),50);
    for(count=0;ctmp[count];count++) ctmp[count] = ctmp[count + 1];
    }

strcat(tmp,ctmp);

/* if the user didnt enter any data, make the default data active */

if (tmp[0] == '\0')
    {
    strcpy(tmp,def);
    }

/* Repaint the top two lines of the screen */

MoveCursor(0,1);
if(lines[vlocy-locy+1] != NULL)
    {
    toggle = 0;
    CleartoEOLN();
    if((vlocy-locy+1) >= STARTBUF && (vlocy-locy+1) <= ENDBUF) StartBold();
    if(lines[vlocy-locy+1][COLUMNS - 1] == 32 && wordwrap == 0)
        {
        lines[vlocy-locy+1][COLUMNS - 1] = '>';
        toggle=1;
        }
    printf("\r%s",lines[vlocy-locy+1]);
    if(toggle) lines[vlocy-locy+1][COLUMNS - 1] = 32;
    if((vlocy-locy+1) >= STARTBUF && (vlocy-locy+1) <= ENDBUF) EndBold();
    }
MoveCursor(0,0);
CleartoEOLN();
if(lines[vlocy-locy] != NULL)
    {
    toggle = 0;
    if((vlocy-locy) >= STARTBUF && (vlocy-locy) <= ENDBUF) StartBold();
    if(lines[vlocy-locy][COLUMNS - 1] == 32 && wordwrap == 0)
        {
        lines[vlocy-locy][COLUMNS - 1] = '>';
        toggle=1;
        }
    printf("%s",lines[vlocy-locy]);
    if(toggle) lines[vlocy-locy][COLUMNS - 1] = 32;
    if((vlocy-locy) >= STARTBUF && (vlocy-locy) <= ENDBUF) EndBold();
    }
MoveCursor(locx,locy);
}

/* This function will clear line x and re-paint it from the array line */

void RedrawLine(int locx, int locy, int vlocy)
{
extern char *lines[];
register int count;
int toggle;
toggle = 0;

if(vlocy>FILELEN) return;

MoveCursor(0,locy);
CleartoEOLN();
if(vlocy >= STARTBUF && vlocy <= ENDBUF) StartBold();
if(vlocy == STARTBUF || vlocy == ENDBUF) StartBold();

if(lines[vlocy][COLUMNS - 1] == 32 && wordwrap == 0)
    {
    lines[vlocy][COLUMNS - 1] = '>';
    toggle=1;
    }

printf("%s",lines[vlocy]);
if(toggle) lines[vlocy][COLUMNS - 1] = 32;
if(vlocy == STARTBUF || vlocy == ENDBUF) EndBold();
if(vlocy >= STARTBUF && vlocy <= ENDBUF) EndBold();
MoveCursor(locx,locy);
}

/* Find out what this terminal is named */

void keyfilename(char tname[])
{
strcpy(tname,"cekm.");
if (getenv("TERM") == NULL)
    {
    strcat(tname,"unknown");
    return;
    }
if (strncat(tname, getenv("TERM"),19) == NULL)
    {
    strcat(tname,"unknown");
    return;
    }
}

/* Create the special-key table and save it out to disk

   ! NOTE ! this routine was quickly hacked together, and isnt
   the nicest code you will see, but it works! :) */

/* To add another special sequence, add its name to the end of
   the table array.  Then go into ce.h and add a DEFINE for it
   being sequentially beyond the last one in the table array.
   Up NUMSEQ by one.  That should be all there is to add sequences... */

void keyset(int killit)
{
struct stat buf;
FILE *ofp;

/* Some special keys the user will press */

char fnam[20];
char fname[256];
char home[100];
int ret;
register int count,count1;
static int seqtable[MAXNUMSEQ][NUMELM];
char tmp;

/* Find out this terminal's name so we can create the file */

if (strncpy(home, getenv("HOME"),100) == NULL) strcpy(home,".");
keyfilename(fnam);
strcat(home,"/.ce");
strcpy(fname,home);
strcat(fname,"/");
strcat(fname,fnam);

/* Get into raw mode, so we can read the keys directly */

Raw(1);

printf("\n\rCreating keymap for %s - carefully answer\n\r",fnam);
printf("the following questions to create a new keymap.\n\r");
printf("Nothing will show as you are typing.\n\r\n\r");

/* Initialize the special-key table to zero's */

for(count=0;count<MAXNUMSEQ;count++)
    for(count1=0;count1<NUMELM;count1++)
        seqtable[count][count1]=0;

/* We need to know what this computer thinks return is, so we
   can check for all other special keys, as well as check for
   return in the main editor */

printf("Please press <return>");

seqtable[CRLF][1] = getchar();
if(seqtable[CRLF][1] == 27) seqtable[CRLF][1] = getchar();
ret=seqtable[CRLF][1];
printf("\r\n");

/* Loop through the special keys defined in table and prompt for
   their input */

printf("Press Up Arrow and then [RETURN] :");
getseq(seqtable[UP],ret);
printf("\n\r");

printf("Press Down Arrow and then [RETURN] :");
getseq(seqtable[DOWN],ret);
printf("\n\r");

printf("Press Left Arrow and then [RETURN] :");
getseq(seqtable[LEFT],ret);
printf("\n\r");

printf("Press Right Arrow and then [RETURN] :");
getseq(seqtable[RIGHT],ret);
printf("\n\r");

printf("Press Backspace and then [RETURN] :");
getseq(seqtable[BACKSPACE],ret);
printf("\n\r");

printf("Press Delete and then [RETURN] :");
getseq(seqtable[DEL],ret);
printf("\n\r");

printf("Press Tab and then [RETURN] :");
getseq(seqtable[TAB],ret);
printf("\n\r");

printf("Press Reverse Tab and then [RETURN] :");
getseq(seqtable[RTAB],ret);
printf("\n\r");

printf("Press Page Up and then [RETURN] :");
getseq(seqtable[PGUP],ret);
printf("\n\r");

printf("Press Page Down and then [RETURN] :");
getseq(seqtable[PGDOWN],ret);
printf("\n\r");

printf("Press Home and then [RETURN] :");
getseq(seqtable[HOME],ret);
printf("\n\r");

printf("Press End and then [RETURN] :");
getseq(seqtable[END],ret);
printf("\n\r");

printf("Press Insert and then [RETURN] :");
getseq(seqtable[INSERT],ret);
printf("\n\r");

printf("\n\rIf you do not have function keys just press return for these\n\r");
printf("\n\r");

/* Loop through 12 function keys */

for(count=4,tmp=1;count<16;count++,tmp++)
    {
    printf("Press F%d and then [RETURN] :",tmp);
    getseq(seqtable[count],ret);
    printf("\n\r");
    }

ret=lstat(home,&buf);

if(ret != 0) mkdir(home, 0700);

if(killit == 0) ofp=fopen(fname,"a");
else            ofp=fopen(fname,"w");

if (ofp == NULL)
    {
    printf("\r\nI cant create the keymap file %s!\r\n",fname);
    return;
    }

for (count=0;count<MAXNUMSEQ;count++)
    {
    for(count1=1;count1<NUMELM;count1++)
        {
        if(seqtable[count][count1] != 0) break;
        }

    if(count1 == NUMELM) continue;

    fprintf(ofp,"%d ",count);

    for(count1=1;count1<NUMELM;count1++)
        {
        fprintf(ofp,"%d ",seqtable[count][count1]);
        }

    fprintf(ofp,"\n");
    }
fclose(ofp);
}

/* The low-level getkey stuff for the Keyset() function */

void getseq(int seq[], int ret)
{
int count, tmp;

for(tmp=getchar(),count=1;count<NUMELM && tmp != ret;tmp=getchar(),count++)
    {
    seq[count] = tmp;
    }
}

void copy(char fname[90], char nname[90])
{
char tmp[300];
FILE *ifp,*ofp;

ifp = fopen(fname,"r");

if (ifp == NULL) return;

ofp = fopen(nname,"w");
if (ofp == NULL) return;

while(fgets(tmp,300,ifp) != NULL)
    {
    fprintf(ofp,"%s",tmp);
    }

fclose(ifp);
fclose(ofp);
}

void search(char string[],int *locy,int *locx, int *vlocy)
{
static int lasty=0;
static int lastx=0;
static char laststr[1000];
int xloc;
int start;
register int count, count1;
char *point;
extern char *lines[];

if (string[0] == '\0') return;
if (*vlocy > FILELEN) start=FILELEN;
else start = *vlocy;

for(count=start;count<=FILELEN;count++)
    {
    point=strstr(lines[count],string);
    if(point == NULL) continue;
    xloc=point - &lines[count][0];
    if(count == lasty && xloc == lastx) continue;
    lastx = *locx = xloc;
    lasty = count;
    gotoline(count+1, locy, locx, vlocy, lines);
    strcpy(laststr,string);
    return;
    }

for(count=0;count<start;count++)
    {
    point=strstr(lines[count],string);
    if(point == NULL) continue;
    xloc=point - &lines[count][0];
    if(count == lasty && xloc == lastx) continue;
    lastx = *locx = xloc;
    lasty = count;
    gotoline(count+1, locy, locx, vlocy, lines);
    strcpy(laststr,string);
    return;
    }

if (strcmp(laststr,string) == 0)
    {
    *locx = lastx;
    gotoline(lasty+1, locy, locx, vlocy, lines);
    strcpy(laststr,string);
    return;
    }
return;
}

void scrolldown(int *locx, int *locy, int *vlocy, char *lines[])
{
int toggle;
toggle=0;

*locy=*locy-1;

if(G_noscroll == 1)
    {
    *vlocy = *vlocy - 1;

    if(*vlocy-*locy+LINES == FILELEN)
        {
        putchar('\a');
        return;
        }

    *vlocy=*vlocy+LINES-*locy;

    if(*vlocy > FILELEN)
        {
        *vlocy=FILELEN-LINES;
        *locy=LINES;
        if(*vlocy < 0) *vlocy=0;
        putchar('\a');
        }
    else
        {
        *locy=1;
        }

    if(MARKING)
        {
        ENDBUF=*vlocy+*locy;
        }

    strcpy(STATUSMSG,DEFAULT_STATUSMSG);

    RedrawScreen(*vlocy,lines,*locx,*locy);

    *vlocy = *vlocy + *locy;

    return;
    }
else
    {
    MoveCursor(0,0);
    DeleteLine();
    MoveCursor(0,*locy);
    InsertLine();

    if (lines[(*vlocy)] != NULL && (*vlocy) <= FILELEN)
        {
        if(*vlocy >=STARTBUF && *vlocy <= ENDBUF) StartBold();
        if(lines[*vlocy][COLUMNS - 1] == 32 && wordwrap == 0)
            {
            lines[*vlocy][COLUMNS - 1] = '>';
            toggle=1;
            }
        printf("%s",lines[*vlocy]);
        if(toggle) lines[*vlocy][COLUMNS - 1] = 32;
        if(*vlocy >=STARTBUF && *vlocy <= ENDBUF) EndBold();
        }
    MoveCursor(*locx,*locy);
    }
}

void scrollup(int *locx, int *locy, int *vlocy, char *lines[])
{
int toggle;
toggle=0;

*locy=*locy+1;

if(G_noscroll == 1)
    {
    *vlocy = *vlocy + 1;

    if(*vlocy-*locy == 0)
        {
        putchar('\a');
        return;
        }

    *vlocy=*vlocy-*locy-LINES;

    if(*vlocy < 0)
        {
        *vlocy = 0;
        putchar('\a');
        }

    strcpy(STATUSMSG,DEFAULT_STATUSMSG);

    *locy = LINES-1;

    if(MARKING)
        {
        ENDBUF=*vlocy+*locy;
        }

    RedrawScreen(*vlocy,lines,*locx,*locy);

    *vlocy = *vlocy + *locy;

    return;
    }
else
    {
    if (*vlocy < 0)
        {
        *vlocy=0;
        }
    else
        {
        MoveCursor(0,LINES);
        DeleteLine();
        MoveCursor(0,0);
        InsertLine();
        if (lines[(*vlocy)] != NULL)
            {
            if(*vlocy >=STARTBUF && *vlocy <= ENDBUF) StartBold();
            if(lines[*vlocy][COLUMNS - 1] == 32 && wordwrap == 0)
                {
                lines[*vlocy][COLUMNS - 1] = '>';
                toggle=1;
                }
            printf("%s",lines[*vlocy]);
            if(toggle) lines[*vlocy][COLUMNS - 1] = 32;
            if(*vlocy >=STARTBUF && *vlocy <= ENDBUF) EndBold();
            }
        MoveCursor(*locx,*locy);
        }
    }
}


/* Map ESC 1-9, 10-12 keys to FKEY 1-12 */

void remap(int *cha, int *func)
{
if ( *func == PRINTABLE && *cha >= '1' && *cha <= '9' )
    {
    *func = FKEY1 + (*cha - '1');
    *cha = 0;
    }
if (*func == PRINTABLE && *cha == '-') *func=FKEY11;
if (*func == PRINTABLE && *cha == '=') *func=FKEY12;
if (*func == PRINTABLE && *cha == '0') *func=FKEY10;
}


void drawstatus(int locx, int locy)
{
extern char **minihelp;
int count;
int indent;
char outbuf[100];

indent = ((55-strlen(STATUSMSG))/2)-1;

for(count=0;count<indent;count++) outbuf[count] = ' ';
outbuf[count] = '\0';
strcat(outbuf,STATUSMSG);

sprintf(STATUSLINE,"CE %-10.10s%-55.55s[%-9.9s]",
        VERSION,outbuf,STATUSMODE);

switch(G_mini)
    {
    case 0:
        {
        LINES++;
        MoveCursor(0,LINES);
        StartBold();
        printf("%s",STATUSLINE);
        EndBold();
        LINES--;
        break;
        }

    case 1:
        {
        LINES++;
        MoveCursor(0,LINES);
        StartBold();
        printf("%s",STATUSLINE);
        EndBold();
        for(count=0;count<MINILINES;count++)
            {
            LINES++;
            MoveCursor(0,LINES);
            if(minihelp[count] != NULL) printf("%s",minihelp[count]);
            else CleartoEOLN();
            }
        LINES-=MINILINES;
        LINES--;
        break;
        }

    case 2:
        {
        for(count=0;count<MINILINES;count++)
            {
            LINES++;
            MoveCursor(0,LINES);
            if(minihelp[count] != NULL) printf("%s",minihelp[count]);
            else CleartoEOLN();
            }
        LINES++;
        MoveCursor(0,LINES);
        StartBold();
        printf("%s",STATUSLINE);
        EndBold();
        LINES--;
        LINES-=MINILINES;
        break;
        }
    }
MoveCursor(locx,locy);
}

#ifdef NORINDEX

char *rindex(char *s, char c)
{
int count;
for(count=strlen(s)-1;count>=0;count--)
    {
    if(s[count] == c) return &s[count];
    }
return NULL;
}

char *index(char *s, char c)
{
int count;
for(count=0;s[count];count++)
    {
    if(s[count] == c) return &s[count];
    }
return NULL;
}

#endif

#ifdef NOSTRSTR

char *strstr(char *s1, char *s2)
{
int count,  count1;
int flag=0;
int nmatch=0;

for(count1=0;s1[count1];count1++)
    {
    for(count=0;s2[count];count++)
        {
        if(s2[count] != s1[count1])
            {
            if(flag==1)
                {
                flag=0;
                count=-1;
                nmatch=0;
                continue;
                }
            nmatch=0;
            count=0;
            break;
            }
        flag=1;
        nmatch++;
        count1++;
        if(nmatch == strlen(s2)) return &s1[count1-nmatch];
        }
    }
return NULL;
}

#endif

void savefile(char *fname, int wordwrap, int quit, int nobaks)
{
extern int vlocy, locy, locx;
extern char *lines[MAXLINES];
FILE *ofp;
char ctmp[101];
int count;
int lastperc;

if(quit)
    {
    Raw(0);
    ClearScreen();
    printf("CE %s Copyright (C) 1991-92, Charles R. Henrich, All Rights "
           "Reserved.\n\r\n\r",VERSION);
    printf("Saving....One Moment Please...");
    fflush(stdout);
    }

if(nobaks == 0)
    {
    strcpy(ctmp, fname);
    strcat(ctmp,".bak");
    copy(fname, ctmp);
    }

ofp=fopen(fname,"w");

if(ofp == 0)
    {
    strcpy(STATUSMSG,"Unable to save file!");
    Raw(1);
    RedrawScreen(vlocy-locy,lines,locx,locy);
    return;
    }

for(count=0;count<=FILELEN;count++)
    {
    if(!wordwrap)
        {
        if(strlen(lines[count])>COLUMNS - 1)
            {
            fprintf(ofp,"%s",lines[count]);
            continue;
            }
        }
    else
        {
        strip(lines,count,wordwrap);
        }

    fprintf(ofp,"%s\n", lines[count]);
    }
fclose(ofp);

if(!quit) return;

putchar('\r');
CleartoEOLN();
printf("Save Complete.\n\r\n\r",fname);
clean_exit(0);
}

int gotoline(int linenum, int *locy, int *locx, int *vlocy, char *lines[])
{
int ret;

linenum--;
if (linenum <0 )
    {
    putchar('\a');
    return -1;
    }

if (linenum > FILELEN) linenum=FILELEN;

if (FILELEN == -1) linenum=0;
if (linenum < *vlocy-*locy || linenum > *vlocy-*locy+LINES)
    {
    ret=2;
    if (linenum > LINES/2)
        {
        *locy=LINES/2;
        *vlocy=linenum;
        if(*vlocy > STARTBUF && MARKING) ENDBUF = *vlocy;
        if(*vlocy < STARTBUF && MARKING) ENDBUF = STARTBUF;
        RedrawScreen(linenum-LINES/2,lines, *locx, *locy);
        }
    else
        {
        *locy=linenum;
        *vlocy=linenum;
        if(*vlocy > STARTBUF && MARKING) ENDBUF = *vlocy;
        if(*vlocy < STARTBUF && MARKING) ENDBUF = STARTBUF;
        RedrawScreen(0,lines, *locx, *locy);
        }
    }
else
    {
    ret=1;
    *locy=linenum-*vlocy+*locy;
    *vlocy=linenum;
    if(*vlocy > STARTBUF && MARKING) ENDBUF = *vlocy;
    if(*vlocy < STARTBUF && MARKING) ENDBUF = STARTBUF;
    if(MARKING) RedrawScreen(*vlocy-*locy ,lines,*locx, *locy);
    MoveCursor(*locx,*locy);
    }
return ret;
}

void readrc(int *insertmode,int *tabsize,int *lcol,
     char *SHELL1, char *SHELL2, int *wordwrap, int *nobaks, int *G_mini,
     int *G_noscroll)
{
struct stat buf;
FILE *ifp;
char name[200];
char tmp[85];
int ret;
int value;

if (strncpy(name, getenv("HOME"),100) == NULL) strcpy(name,".");

strcat(name,"/.ce/cerc");

ifp=fopen(name,"r");

if(ifp == NULL)
    {
    strcpy(name,CONFIG_PATH);
    strcat(name,"cerc");
    ifp=fopen(name,"r");
    if(ifp == NULL) return;
    }

while(fgets(tmp,81,ifp) != NULL)
    {
    if(tmp[0] == '#') continue;
    if(sscanf(tmp,"%s",name)!=1) continue;
    if(strcmp("shell1",name) == 0)
        {
        ret=sscanf(tmp,"%[^\"]\"%[^\"]\"",
            name,name);
        if(ret != 2) continue;
        strcpy(SHELL1,name);
        continue;
        }
    if(strcmp("shell2",name) == 0)
        {
        ret=sscanf(tmp,"%[^\"]\"%[^\"]\"",
            name,name);
        if(ret != 2) continue;
        strcpy(SHELL2,name);
        continue;
        }
    if(sscanf(tmp,"%s %d",name,&value)!=2) continue;

    if(strcmp("insert_mode",name) == 0)
        {
        if(value != 0 && value != 1) continue;
        *insertmode=value;
        continue;
        }

    if(strcmp("no_scroll",name) == 0)
        {
        if(value != 0 && value != 1) continue;
        *G_noscroll=value;
        continue;
        }

    if(strcmp("tab_space",name) == 0)
        {
        *tabsize=value;
        continue;
        }
    if(strcmp("left_margin",name) == 0)
        {
        *lcol = value - 1;
        continue;
        }
    if(strcmp("mini_help",name) == 0)
        {
        if(value != 0 && value != 1) continue;
        *G_mini = value;
        continue;
        }
    if(strcmp("word_wrap",name) == 0)
        {
        if(value != 0 && value != 1) continue;
        *wordwrap=value;
        continue;
        }
    if(strcmp("no_bak_files",name) == 0)
        {
        if(value != 0 && value != 1) continue;
        *nobaks=value;
        continue;
        }
    }
fclose(ifp);
return;
}

char *getstr(int locx, int locy,char *string)
{
int startx;
int cha,func;
int term;
int count;
int begin=1;

startx = locx;
locx = cha = func = term =0;

while(1)
    {
    getkey(&cha, &func);

    if(begin)
        {
        CleartoEOLN();
        begin=0;
        }

    switch(func)
        {
        case CONTROL:
            switch(cha)
                {
                case 'D':
                    string[0]='\0';
                    locx=0;
                    MoveCursor(locx+startx,locy);
                    CleartoEOLN();
                    MoveCursor(locx+startx,locy);
                    break;

                case 'C':
                    locx=strlen(string);
                    MoveCursor(locx+startx,locy);
                    break;

                case 'Z':
                    locx=0;
                    MoveCursor(locx+startx,locy);
                    break;

                case 'J':
                case 'M':
                    term=1;
                    break;

                }
            break;

        case DEL:
            if(deletechars(1, locx, string) == 1)
                {
                for(count=locx;string[count];count++)
                    {
                    putchar(string[count]);
                    }
                CleartoEOLN();
                MoveCursor(locx+startx,locy);
                }
            break;

        case CRLF:
            term=1;
            break;

        case LEFT:
            if(locx == 0) break;
            locx--;
            MoveCursor(locx+startx,locy);
            break;

        case RIGHT:
            if(locx == COLUMNS-1) break;
            locx++;
            MoveCursor(locx+startx,locy);
            break;

        case BACKSPACE:
            locx--;
            if(locx < 0) locx = 0;
            MoveCursor(locx+startx,locy);
            if(deletechars(1, locx, string) == 1)
                {
                for(count=locx;string[count];count++)
                    {
                    putchar(string[count]);
                    }
                CleartoEOLN();
                MoveCursor(locx+startx,locy);
                }
            break;

        case PRINTABLE:
            if(locx > strlen(string))
                {
                padchars(strlen(string),locx,string);
                }
            if(insertchars(1,locx,cha,string) == 1)
                {
                for(count=locx;string[count];count++)
                    {
                    putchar(string[count]);
                    }
                locx++;
                }
            else
                {
                putchar('\a');
                break;
                }
            MoveCursor(locx+startx,locy);
            break;

        default:
            break;
        }
    if(term == 1) break;
    }
return string;
}

int wrapline(int locx, int locy, int vlocy, char *lines[], int lcol, int *ret)
{
char ctmp[90];
char linebuf[90];
char *ptmp;
int nextlineblank;
int nonextline;
int startx;
int count;

ctmp[0] = linebuf[0] = '\0';
ptmp = NULL;
nextlineblank = 1;
nonextline = 0;
*ret = 0;

if (vlocy >= MAXLINE) return -1;
if (FILELEN+1 > MAXLINE) return -1;

if (vlocy > FILELEN)
    {
    padlines(vlocy, lines);
    }

ptmp = (char *) rindex(lines[vlocy],' ');

if(ptmp == NULL) return -1;

*ptmp = '\0';

ptmp++;

startx = ptmp - &lines[vlocy][0];

strcpy(linebuf,ptmp);

if(strlen(linebuf) < (COLUMNS - lcol))
    {
    for(count=startx;count<CALLOCSIZE;count++)
        {
        lines[vlocy][count] = '\0';
        }

    MoveCursor(startx,locy);

    vlocy++;
    locy++;

    if(vlocy > FILELEN) nonextline = 1;

    if(nonextline == 0)
        {
        for(count=0;lines[vlocy][count];count++)
            {
            if(!isspace(lines[vlocy][count]))
                {
                nextlineblank = 0;
                break;
                }
            }
        }

    if((nonextline == 1) || (nextlineblank == 1) ||
       (strlen(linebuf) + strlen(lines[vlocy])+lcol+1 > COLUMNS-1))
        {
        insertlines(lines, vlocy, 1);
        ctmp[0]='\0';
        for(count=0;count<lcol;count++) ctmp[count] = ' ';
        ctmp[count]='\0';
        lines[vlocy][0]='\0';

        if(strlen(linebuf) + lcol < COLUMNS)
            {
            strcat(lines[vlocy],ctmp);
            }

        strcat(lines[vlocy],linebuf);
        *ret=1;
        }
    else
        {
        if(strlen(lines[vlocy]) < lcol)
             padchars(strlen(lines[vlocy]), lcol, lines[vlocy]);

        insertchars(strlen(linebuf)+1, lcol, ' ', lines[vlocy]);

        for(count=lcol;count<strlen(linebuf)+lcol;count++)
            {
            lines[vlocy][count] = linebuf[count-lcol];
            }
        *ret=2;
        }
    }
else
    {
    return -1;
    }

return startx;
}

void clean_exit(int return_code)
{
int l,c;

ScreenSize(&l,&c);
SetScroll(0,l);
MoveCursor(0,4);
Raw(0);
signal(SIGHUP,SIG_IGN);
exit(return_code);
}

char *itoa(int var)
{
static char outstr[50];

sprintf(outstr,"%d",var);

return outstr;
}
