/*****************************************************************************
* CE v1.2 Copyright (C) 1991-92, Charles R. Henrich, All Rights Reserved.    *
*                                                                            *
* Use in source and binary forms, with or without modification and           *
* Redistribution without modification are permitted provided that the        *
* following conditions are met:                                              *
*                                                                            *
* 1. Redistributions of source code must retain the above copyright          *
*    notice, all internal copyright notices, this list of conditions         *
*    and the following disclaimer.                                           *
* 2. Redistributions in binary form must reproduce the above copyright       *
*    notice, this list of conditions and the following disclaimer in the     *
*    documentation and/or other materials provided with the distribution.    *
* 3. All advertising materials mentioning features or use of this software   *
*    must display the following acknowledgement:                             *
*      CE Was developed in 1991-92 by Charles Henrich                        *
* 4. Neither the name of this product (CE) nor the name of the author        *
*    (Charles Henrich) may be used to endorse products that include this     *
*    product without specific prior written permission.                      *
* 5. No profit may be made from the inclusion, or distribution of this       *
*    product.  Although a reasonable service charge for the distribution     *
*    of this product is permitted where appropriate.                         *
*                                                                            *
* THIS SOFTWARE IS PROVIDED BY CHARLES HENRICH `` AS IS '' AND ANY EXPRESS   *
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED          *
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE     *
* DISCLAIMED.  IN NO EVENT SHALL CHARLES HENRICH BE LIABLE FOR ANY DIRECT,   *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES         *
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR         *
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)         *
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT *
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY  *
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF     *
* SUCH DAMAGE.                                                               *
*                                                                            *
******************************************************************************
*                                                                            *
* The file cemain.c contains the main handling of input and output of key    *
* presses, and their actions.  Also included in this file is the functions   *
* to deal with signal catching.                                              *
*                                                                            *
*****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>

#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>
#include "ce.h"
#ifndef NO_STDLIB
#include <stdlib.h>
#endif

/****************************************************************************/
/*                                                                          */
/* Global Variables (need to change them all to G_<lowercase> format        */
/* eventually.                                                              */
/*                                                                          */
/****************************************************************************/

int wordwrap;                            /* 1 if wordwrap is on, 0 if off   */
int MARKING;                             /* 1 if in marking, 0 if not       */
int COLUMNS;                             /* Columns on screen, starts at 1  */
int LINES;                               /* Lines on screen, starts at 1    */
int MAXLINE=MAXLINES;                    /* Max num of lines allowed        */
int FILELEN;                             /* Current length of file starts 0 */
int BUFLEN;                              /* Current buffer length starts 0  */
int HELPLINES;                           /* Number of lines in help file    */
int STARTBUF,ENDBUF;                     /* Start and end line for buffer   */

char STATUSLINE[90];                     /* Status line                     */
char STATUSMODE[10]="OVERWRITE";         /* Status mode                     */
char STATUSMSG[46]=DEFAULT_STATUSMSG;    /* Status message                  */
char SHELL1[200]="spell %s";             /* Shell escape #1 variable        */
char SHELL2[200]="cc %s -lm -o %s -O";   /* Shell escape #2 variable        */

char *help[MAXLINES];                    /* Pointer array to help text      */
char *lines[MAXLINES];                   /* Pointer array to body of text   */
char *buffer[MAXLINES];                  /* Pointer array to buffer space   */

int locx;                                /* X location of cursor starts 0   */
int locy;                                /* Y location of cursor starts 0   */
int vlocy;                               /* Line of location in file st. 0  */
int tabsize;                             /* Number of spaces for a tab      */
int G_mini;                              /* 1 if minihelp is on, 0 if off   */
int G_noscroll;                          /* 1 if we dont scroll 0 if we do  */

extern char **helppoint;                 /* Pointer to help text            */

void hangup();
void winchange();
void majerr();
void dumpfile();

int main(int argc, char *argv[])
{
struct stat buf;                         /* Structure for stat function     */
register int count,count1;               /* Counting variables              */
int cha;                                 /* Character the user pressed      */
int func;                                /* Function the user pressed       */
int ret;                                 /* Genric return value             */
int cret;                                /* Return from cursor movement     */
int tmp;                                 /* Temp integer var                */
int tmp1;                                /* Temp integer var                */
int tmp2;                                /* Temp integer var                */
int insertmode;                          /* 1 if in insertmode, 0 if not    */
int pusage;                              /* 1 if error in command line synx */
int lcol;                                /* Left column (starts at 0)       */
int nobaks;                              /* 1 if we arent to create .baks   */
int lastdel;                             /* Number of chars in last ^E req  */
int startline;                           /* Line number to start cursor on  */
char *ptmp;                              /* Temp char pointer var           */
char ctmp1;                              /* Temp char var                   */
char ctmp[90];                           /* Temp string variable            */
char ctmp2[90];                          /* Temp string variable            */
char fname[90];                          /* Filename we are editing         */
char deletebuf[90];                      /* Buffer for a line delete        */
char linebuf[90];                        /* Generic line buffer             */
char lastsearch[90];                     /* Last search string              */

FILE *ifp;                               /* Input file pointer              */
FILE *ofp;                               /* Output file pointer             */

if(argc > 7)
    {
    printf("\a\nCE Usage: CE (filename/-keyset) [-w -m -s -h<helpfile>");
    printf(" -l<line#>]\n\n");
    exit(-10);
    }

/* Catch Hangups, And all other possible signals that would
   be VERY bad for a person in the middle for an edit! */

/* On a hangup signal, execute the hangup() function */

signal(SIGHUP, hangup);
signal(SIGINT, SIG_IGN);
signal(SIGQUIT, SIG_IGN);
signal(SIGTERM, SIG_IGN);
#ifdef SIGDANGER
signal(SIGDANGER, dumpfile);
#endif
#ifdef SIGWINCH
signal(SIGWINCH, winchange);
#endif

/* Handle the *really* nasty bugs nicely */

signal(SIGBUS, majerr);
signal(SIGSEGV, majerr);

/******************************************/
/*                                        */
/* Initialize all the variables to the    */
/* begin state, mostly zero's and nulls   */
/*                                        */
/******************************************/

for(count=0;count<MAXLINE;count++)
    {
    lines[count]=NULL;
    }

for(count=0;count<90;count++)
    {
    linebuf[count]='\0';
    ctmp[count]='\0';
    deletebuf[count]='\0';
    fname[count]='\0';
    STATUSLINE[count]='\0';
    lastsearch[count]='\0';
    }

/* Set the MAXLINE of the editor to sync with the arrays, 0 to n-1 */

MAXLINE--;

HELPLINES = MARKING = FILELEN=0;
STARTBUF = ENDBUF = BUFLEN=-1;

count = count1 = cha = func = locx = locy = vlocy = cret = ret = 0;
startline = nobaks = pusage = tmp = tmp1 = insertmode = lcol = tmp2 = 0;
lastdel = 90;
ctmp1 = ' ';
G_noscroll = G_mini = 0;
wordwrap = 1;
tabsize = 4;

/******************************************/
/*                                        */
/* Call the help routines to initialize   */
/* the beginning help arrays.             */
/*                                        */
/******************************************/

setup_help();

/******************************************/
/*                                        */
/* Read all the users default settings    */
/* from the .cerc file.                   */
/*                                        */
/******************************************/

readrc(&insertmode,&tabsize,&lcol, SHELL1, SHELL2,
    &wordwrap, &nobaks, &G_mini, &G_noscroll);

/* Set the beginning mode display */

if(insertmode) strcpy(STATUSMODE,"INSERT");

/* Default STATUSLINE */

sprintf(STATUSLINE,"CE %-10s    %-45s     [%-9s]",
        VERSION,STATUSMSG,STATUSMODE);

/* We set up the size of the screen here, must be only 80 columns
The Bottom line of the screen is also removed, which saves us the trouble
of trying to figure out if this particular terminal does what at the
right margin on the bottom line.  Clear the screen and draw the contents
of memory onto the screen */

ret=InitScreen();                        /* Initialize the curses routines */

ScreenSize(&LINES,&COLUMNS);             /* Get the current screen size */

if (COLUMNS < 80)
    {
    ClearScreen();
    printf("\a\aTerminal MUST support at least 80 columns!\n\r\n\r");
    clean_exit(-4);
    }

if (COLUMNS > 80) COLUMNS=80;

/******************************************/
/*                                        */
/* If the minihelp is set to on, we have  */
/* to take the bottom three lines of the  */
/* screen to ourself, otherwise just take */
/* one.                                   */
/*                                        */
/******************************************/

if(G_mini == 0) LINES--;
if(G_mini == 1 || G_mini == 2) LINES-=(MINILINES+1);

/* GIVE US SOMETHING TO TYPE IN */

padlines(0, lines);

/******************************************/
/*                                        */
/* Parse the command line if neccesary,   */
/* at this point we also read in the      */
/* alternate help screen if present.      */
/*                                        */
/******************************************/

if (argc > 1)
    {
    for(count=1,pusage=0;count<argc;count++)   /* loop through the options */
        {
        if (strcmp(argv[count],"-keyset") == 0)   /* Do keyset then exit */
            {
            keyset(0);
            ClearScreen();
            printf("Keymap created\r\n");
            clean_exit(0);
            }

        if (strncmp(argv[count],"-l",2) == 0 ||
            strncmp(argv[count],"-L",2) == 0)
            {
            if(&argv[count][2] != NULL) startline=atoi(&argv[count][2]);
            continue;
            }

/******************************************/
/*                                        */
/* If the user needs the mini menu on,    */
/* subtract the proper number of lines    */
/* from the screen.                       */
/*                                        */
/******************************************/

        if (strcmp(argv[count],"-m") == 0 || strcmp(argv[count],"-M") == 0)
            {
            if(G_mini == 0)
                {
                G_mini = 1;
                LINES-=MINILINES;
                }
            else
                {
                if(G_mini == 1)
                    {
                    G_mini = 0;
                    LINES+=MINILINES;
                    }
                }
            continue;
            }

/******************************************/
/*                                        */
/* Alternative Help?  If so read it into  */
/* the help array, and set the HELPLINES  */
/* to the number of lines read in.        */
/*                                        */
/******************************************/

        if (strncmp(argv[count],"-h",2) == 0 ||
            strncmp(argv[count],"-H",2) == 0)
            {

            ifp = fopen(&(argv[count][2]), "r");

            if(ifp != NULL)
                {
                ret=stat(&(argv[count][2]),&buf);

                if(ret == 0)
                    {
                    if(!(S_ISREG(buf.st_mode)))
                        {
                        fclose(ifp);
                        continue;
                        }
                    }

                for(count1=0;count1<=MAXLINE;count1++)
                    {
                    if(fgets(ctmp, COLUMNS+1, ifp) == NULL) break;
                    padlines(count1, help);
                    strcpy(help[count1],ctmp);
                    strip(help,count1,wordwrap);
                    FILELEN = count1;
                    }

                HELPLINES=FILELEN;
                helppoint = help;
                FILELEN=0;
                fclose(ifp);
                }
            continue;
            }

        if (strcmp(argv[count],"-w") == 0 || strcmp(argv[count],"-W") == 0)
            {
            if(wordwrap == 1) wordwrap = 0;
            else wordwrap = 1;
            continue;
            }

        if (strcmp(argv[count],"-s") == 0 || strcmp(argv[count],"-S") == 0)
            {
            if(G_noscroll == 1) G_noscroll = 0;
            else G_noscroll = 1;
            continue;
            }

/******************************************/
/*                                        */
/* If the parameter doesnt begin with a   */
/* dash, we assume its a file name        */
/*                                        */
/******************************************/

        if(argv[count][0] != '-')
            {
            if(fname[0] != '\0') pusage = 1;
            strcpy(fname,argv[count]);
            continue;
            }

        printf("\n\rUnknown option %s\n\r",argv[count]);
        pusage=1;
        }

    if(pusage == 1)
        {
        Raw(0);
        ClearScreen();
    printf("\a\n\rCE Usage: CE (filename/-keyset) [-w -m -s -h<helpfile>");
    printf(" -l<line#>]\n\r\n\r");
        clean_exit(-15);
        }
    }

/******************************************/
/*                                        */
/* If we dont have a filename, either     */
/* preselect the name "cefile" or prompt  */
/* for the name (based on #define)        */
/*                                        */
/******************************************/

if(fname[0] == '\0')
    {

#ifdef FORCE_FILENAME

    printf("Filename to edit [cefile]:");
    fflush(stdout);
    fgets(fname,90,stdin);

    for(count=0;fname[count];count++)
        {
        if(fname[count] == 13 || fname[count] == 10) fname[count] = '\0';
        }

    if(fname[0] == '\0') strcpy(fname,"cefile");

#else

    strcpy(fname,"cefile");

#endif
    }

/******************************************/
/*                                        */
/* Try and open fname for reading, if we  */
/* get an open, the file is there         */
/* (obviously).  Read the file in if it   */
/* is there.                              */
/*                                        */
/******************************************/

ifp = fopen(fname, "r");

if(ifp != NULL)
    {
    ret=stat(fname,&buf);                /* Stat the file */

    if(ret == 0)
        {
        if(!(S_ISREG(buf.st_mode)))      /* Is it a regular file? */
            {                            /* Nope! */
            fclose(ifp);
            ClearScreen();
            printf("\a\r\nCE: %s is not an editable file!\r\n\r\n",fname);
            clean_exit(-21);
            }
        }

/******************************************/
/*                                        */
/* Begin to read in the file, we go from  */
/* 0 to MAXLINE, allocating each 80 bytes */
/* of memory with the padlines routine.   */
/* We then strip that line and increment  */
/* the FILELEN.                           */
/*                                        */
/******************************************/

    for(count=0;count<=MAXLINE;count++)
        {
        if(fgets(ctmp, COLUMNS+1, ifp) == NULL) break;
        padlines(count, lines);
        strcpy(lines[count],ctmp);
        strip(lines,count,wordwrap);
        FILELEN = count;
        }

/******************************************/
/*                                        */
/* If the FILELEN = MAXLEN chances are    */
/* very good we truncated the file, so    */
/* warn the user and pause for a moment   */
/* of silence :)                          */
/*                                        */
/******************************************/

    if(FILELEN == MAXLINE)
        {
        printf("\nWarning!  File truncated at %d lines\a\n",MAXLINES);
        sleep(3);
        }

    fclose(ifp);
    }

/******************************************/
/*                                        */
/* Set raw mode to on, read in the key    */
/* sequences, and draw the first page of  */
/* the screen.  Also attempt to set the   */
/* scrolling region (if the terminal      */
/* supports it)                           */
/*                                        */
/******************************************/

Raw(1);
keyinit();
RedrawScreen(0, lines, locx, locy);
ret=SetScroll(0,LINES);

if(startline != 0) gotoline(startline, &locy, &locx, &vlocy, lines);

/* The main loop of the program, do it forever */

/* The Getkey function returns many possible sequences based
   on what the user types, they can be read in the cefunc.c file */

while(1)
    {
    getkey(&cha,&func);

/******************************************/
/*                                        */
/* In this IF we process all possible ESC */
/* key presses.                           */
/*                                        */
/******************************************/

    if(func == ESC)
        {

/******************************************/
/*                                        */
/* The user hit ESC, and has waited long  */
/* enough to cause a time out, so lets be */
/* nice and prompt the user for the next  */
/* key.                                   */
/*                                        */
/******************************************/

        if(cha != MOREDATA)
            {
            strcpy(STATUSMSG,"Press 1 to exit.  Press h for help");
            drawstatus(locx,locy);
            while(func == ESC) getkey(&cha,&func);
            strcpy(STATUSMSG,DEFAULT_STATUSMSG);
            drawstatus(locx,locy);
            remap(&cha,&func);     /* Map ESC keys to FKEY equivilents */
            }
        else                       /* No timeout */
            {
            getkey(&cha,&func);
            remap(&cha,&func);     /* Map ESC keys to FKEY equivilents */
            }

/******************************************/
/*                                        */
/* The user hit ESC-H lets show them help */
/* we need to make sure that the marking  */
/* begin and ends are set to -1 so we     */
/* dont show any marking on the help      */
/* screens.                               */
/*                                        */
/******************************************/

        if (func == PRINTABLE && cha == 'h')
            {
            strcpy(STATUSMSG,"Press any key to exit help");
            strcpy(STATUSMODE,"HELP");

            tmp = STARTBUF;
            tmp1 = ENDBUF;
            STARTBUF = ENDBUF = -1;

            show_help();

            STARTBUF = tmp;
            ENDBUF = tmp1;

            if(insertmode) strcpy(STATUSMODE,"INSERT");
            else if(MARKING) strcpy(STATUSMODE,"MARKING");
            else strcpy(STATUSMODE,"OVERWRITE");

            strcpy(STATUSMSG,DEFAULT_STATUSMSG);

            RedrawScreen(vlocy-locy,lines,locx,locy);
            continue;
            }

/******************************************/
/*                                        */
/* ESC-S The user wants to save his       */
/* current settings.  Try and open the    */
/* ~/.cerc file for writing, if we fail   */
/* let the user know, and pause else we   */
/* go ahead and dump out the users currnt */
/* settings.                              */
/*                                        */
/******************************************/

        if(func == PRINTABLE && cha == 's')
            {
            if(strncpy(ctmp, getenv("HOME"),90) == NULL) strcpy(fname,".");
            strcat(ctmp,"/.ce");

            ret=lstat(ctmp,&buf);

            if(ret != 0) mkdir(ctmp, 0700);

            strcat(ctmp,"/cerc");

            ifp=fopen(ctmp,"w");

            if(ifp == NULL)
                {
                strcpy(STATUSMSG,"\aUnable to create file");
                drawstatus(locx,locy);
                sleep(3);
                strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                drawstatus(locx,locy);
                continue;
                }

            strcpy(STATUSMSG,"Creating defaults file");
            drawstatus(locx,locy);

            fprintf(ifp,"# 0 means off, 1 means on.\n\n");
            fprintf(ifp,"# shell1 and shell2 define what "
                        "actions will happen on pressing of F6/F7\n");
            fprintf(ifp,"# The first %%s will be filename, while"
                        "the second will be the filename\n");
            fprintf(ifp,"# without the .* extension\n\n");

            fprintf(ifp,"insert_mode    %d\n",insertmode);
            fprintf(ifp,"tab_space      %d\n",tabsize);
            fprintf(ifp,"left_margin    %d\n",lcol+1);
            fprintf(ifp,"word_wrap      %d\n",wordwrap);
            fprintf(ifp,"no_bak_files   %d\n",nobaks);
            fprintf(ifp,"mini_help      %d\n",G_mini);
            fprintf(ifp,"no_scroll      %d\n",G_noscroll);
            fprintf(ifp,"shell1         \"%s\"\n",SHELL1);
            fprintf(ifp,"shell2         \"%s\"\n",SHELL2);
            fclose(ifp);
            sleep(1);
            strcpy(STATUSMSG,DEFAULT_STATUSMSG);
            drawstatus(locx,locy);
            continue;
            }

/******************************************/
/*                                        */
/* Okay, the user wants to create a       */
/* keymap, lets go to it!                 */
/*                                        */
/******************************************/

        if(func == PRINTABLE && cha == 'k')
            {
            ClearScreen();

            ret=0;

            while(ret != 'n' && ret != 'y')
                {
                printf("Should I remove the current keymap file (y/n) ");
                ret=getchar();
                if(ret != 'n' && ret != 'y') printf("\r\n");
                }

            if(ret == 'y')
                {
                printf("Yes\r\n\r\n");
                keyset(1);
                }
            else
                {
                printf("No\r\n\r\n");
                keyset(0);
                }

            keyinit();

            RedrawScreen(vlocy-locy,lines,locx,locy);
            continue;
            }

/******************************************/
/*                                        */
/* The user either wants to turn the mini */
/* menu on or off, if on subtract three   */
/* lines from the screen, or add them.    */
/* We also need to reset the scrolling    */
/* region, and repain the screen.         */
/*                                        */
/******************************************/

        if(func == PRINTABLE && cha == 'm')
            {
            if(G_mini == 0)
                {
                G_mini = 1;
                LINES-=MINILINES;
                }
            else
                {
                if(G_mini == 1)
                    {
                    G_mini = 0;
                    LINES+=MINILINES;
                    }
                }

            if(locy > LINES)
                {
                vlocy = vlocy-locy+LINES;
                locy = LINES;
                }

            SetScroll(0,LINES);
            RedrawScreen(vlocy-locy,lines,locx,locy);
            continue;
            }
        }

/******************************************/
/*                                        */
/* The following mappings occur here,     */
/*                                        */
/* Page Up      ->      Control B         */
/* Page Down    ->      Control F         */
/* Home         ->      Control Z         */
/* End          ->      Control C         */
/* Insert       ->      Control T         */
/*                                        */
/******************************************/

    switch(func)
        {
        case PGUP:
            {
            func=CONTROL;
            cha='B';
            break;
            }

        case PGDOWN:
            {
            func=CONTROL;
            cha='F';
            break;
            }

        case HOME:
            {
            func=CONTROL;
            cha='Z';
            break;
            }

        case END:
            {
            func=CONTROL;
            cha='C';
            break;
            }

        case INSERT:
            {
            func=CONTROL;
            cha='T';
            break;
            }
        }

/******************************************/
/*                                        */
/* Ok, they user hasnt done an ESC key at */
/* this point, we have already remapped   */
/* the keypad, and so now we go onto the  */
/* normal extended functions.             */
/*                                        */
/******************************************/

    switch(func)
        {

/******************************************/
/*                                        */
/* Aint got a clue what the user just did */
/* so lets do our best to ignore it.      */
/*                                        */
/******************************************/

        case UNKNOWN:
            {
            fflush(stdin);
            fflush(stdout);
            break;
            }

/******************************************/
/*                                        */
/* The user has pressed a printable key,  */
/* lets make sure the key is inside the   */
/* right ranges, and do the painting.     */
/*                                        */
/******************************************/

        case PRINTABLE:
            {

/******************************************/
/*                                        */
/* If the current line is > than the      */
/* MAXLINE we dont want do do anything,   */
/* just break.  If the line is > than the */
/* FILELEN lets extend the FILELEN to     */
/* include us (allocate us memory) with   */
/* the padlines function.                 */
/*                                        */
/******************************************/

            if (vlocy > MAXLINE) break;

            if (vlocy > FILELEN)
                {
                padlines(vlocy, lines);
                }

/******************************************/
/*                                        */
/* If the character is at a position >    */
/* than the last character in the string, */
/* pad the string with spaces up until    */
/* this character so we have a place to   */
/* type.                                  */
/*                                        */
/******************************************/

            if (locx > strlen(lines[vlocy]))
                {
                padchars(strlen(lines[vlocy]), locx, lines[vlocy]);
                }

/******************************************/
/*                                        */
/* If we are in insertmode, attempt to    */
/* insert the character at the current    */
/* location.  If it fails, lets try to    */
/* wrap the line, and then attempt the    */
/* insert again, if it fails again, give  */
/* up and beep at the user.               */
/*                                        */
/******************************************/

            if(insertmode)
                {
                if(insertchars(1,locx,cha,lines[vlocy]) == 1)   /* Valid ins */
                    {
                    RedrawFromX(locx, locy, lines[vlocy]); /* Insert the cha */
                    locx++;
                    MoveCursor(locx,locy);
                    }
                else               /* Insert failed, lets wrap to make space */
                    {
                    tmp=0;
                    ret=wrapline(locx, locy, vlocy, lines, lcol, &tmp);

                    if(ret >= 0)                    /* Was it a valid wrap ? */
                        {
                        MoveCursor(ret,locy);
                        if(ret < COLUMNS) CleartoEOLN();

/******************************************/
/*                                        */
/* If this line isnt the one right before */
/* the bottom of the screen, we need to   */
/* update the next line.  If wrapline     */
/* returned tmp to be 1, then wrapline    */
/* has inserted a blank line in the text  */
/* so we need to do the same.  In either  */
/* case we also then need to repaint the  */
/* next line on the screen, to update its */
/* wordwrapped state.                     */
/*                                        */
/******************************************/

                        if(locy < LINES)
                            {
                            if(tmp == 1)
                                {
                                MoveCursor(0,LINES);
                                DeleteLine();
                                MoveCursor(0,locy+1);
                                InsertLine();
                                MoveCursor(locx,locy);
                                }
                            RedrawLine(locx,locy+1,vlocy+1);
                            }

/******************************************/
/*                                        */
/* The wrapline function returns the loc  */
/* of the character it began to wrap on.  */
/* If that character is BEFORE the cursor */
/* position, we need to move the cursor   */
/* to the next line, as well as keeping   */
/* its place in the word.                 */
/*                                        */
/******************************************/

                        if(ret <= locx)     /* We need to move the cursor */
                            {
                            locx = locx - ret+lcol;
                            locy++;
                            vlocy++;
                            }

/******************************************/
/*                                        */
/* Update the cursors screen position,    */
/* then check the return code to see if   */
/* we need to scrolldown or not.          */
/*                                        */
/******************************************/

                        ret=MoveCursor(locx,locy);

                        if(ret == SCROLLDOWN)
                            {
                            scrolldown(&locx,&locy,&vlocy,lines);
                            }

/******************************************/
/*                                        */
/* We have now wrapped the line, so lets  */
/* try to insert a character again, if    */
/* it fails this time, just beep.         */
/*                                        */
/******************************************/

                        if(insertchars(1,locx,cha,lines[vlocy]) == 1)
                            {
                            RedrawFromX(locx, locy, lines[vlocy]);
                            locx++;
                            MoveCursor(locx,locy);
                            }
                        else
                            {
                            putchar('\a');
                            }
                        }

                    else                       /* Wrap failed */
                        {
                        putchar('\a');
                        }
                    }
                }
            else                            /* Not in insertmode, just dump */
                {                           /* the char out there... */
                lines[vlocy][locx] = cha;
                putchar(cha);
                locx++;
                }

/******************************************/
/*                                        */
/* Okay, we are sitting on the rightmost  */
/* edge of the screen, we need to see if  */
/* we should wrap here, if so, and its    */
/* successful, we move the cursor and     */
/* update the screen to reflect the       */
/* wrapping.                              */
/*                                        */
/******************************************/

            if(locx == COLUMNS)
                {
                if(wordwrap == 1)
                    {
                    tmp=0;
                    ret=wrapline(locx, locy, vlocy, lines, lcol, &tmp);

                    if(ret >= 0)                  /* We wrapped.. */
                        {
                        MoveCursor(ret,locy);
                        if(ret < COLUMNS) CleartoEOLN();

                        if(locy < LINES)
                            {
                            if(tmp == 1)
                                {
                                MoveCursor(0,LINES);
                                DeleteLine();
                                MoveCursor(0,locy+1);
                                InsertLine();
                                MoveCursor(locx,locy);
                                }
                            RedrawLine(locx,locy+1,vlocy+1);
                            }

                        if(ret <= locx)
                            {
                            locx = locx - ret+lcol;
                            }
                        else
                            {
                            locx=lcol;
                            }

                        vlocy++;                  /* Move the cursor (mem) */
                        locy++;
                        }

                    ret=MoveCursor(locx,locy);

                    if(ret == SCROLLDOWN)
                        {
                        scrolldown(&locx,&locy,&vlocy,lines);
                        }
                    }
                else
                    {
                    locx = lcol;
                    locy++;
                    vlocy++;
                    cret = MoveCursor(locx,locy);
                    }
                }
            break;
            }                               /* END OF PRINTABLE CASE */


/******************************************/
/*                                        */
/* The user hit the UP arrow key, if we   */
/* are in a marking state, we need to     */
/* move the bottom edge of the marking    */
/* region UP, unless of course we are     */
/* at the start of the marking region.    */
/*                                        */
/* Move the cursor up as well.            */
/*                                        */
/******************************************/

        case UP:
            {
            if(vlocy == 0) break;

            if(MARKING)
                {
                if(vlocy > STARTBUF)
                    {
                    ENDBUF=vlocy-1;
                    RedrawLine(locx,locy,vlocy);
                    }
                }

            locy--;
            vlocy--;
            cret=MoveCursor(locx,locy);
            break;
            }

/******************************************/
/*                                        */
/* The user has pressed the down arrow    */
/* key, if we are marking and we are in-  */
/* side of the marking region we need to  */
/* extend it one line.  We also check     */
/* here to make sure the user is within   */
/* the MAXLINE's boundry.                 */
/*                                        */
/******************************************/

        case DOWN:
            {
            if(vlocy >= MAXLINE) break;

            if(MARKING && vlocy >= FILELEN) break;

            locy++;
            vlocy++;

            if(MARKING)
                {
                if(vlocy > STARTBUF)
                    {
                    ENDBUF=vlocy;
                    if(locy <= LINES) RedrawLine(locx,locy,vlocy);
                    }
                }

            cret=MoveCursor(locx,locy);
            break;
            }

/******************************************/
/*                                        */
/* Tab has a "mode" type thing to it,     */
/* if the user is in overwrite, just move */
/* the cursor tabsize spaces to the right */
/* If in insertmode, move the cursor      */
/* tabsize spaces to the right, inserting */
/* blanks for each space.  If a marking   */
/* region is set, insert tabsize spaces   */
/* at the beginning of each line in the   */
/* block.                                 */
/*                                        */
/******************************************/

        case TAB:
            {
/******************************************/
/*                                        */
/* If the lines of the block are visible  */
/* Redraw them after we insert into them. */
/*                                        */
/******************************************/

            if(MARKING == 1)
                {
                for(count=STARTBUF;count<=ENDBUF;count++)
                    {
                    strip(lines,count,wordwrap);
                    insertchars(tabsize,0,' ',lines[count]);
                    if(count >= vlocy-locy && count <= vlocy-locy+LINES)
                        {
                        RedrawLine(0,count-(vlocy-locy),count);
                        }
                    }
                MoveCursor(locx,locy);
                break;
                }

            if(insertmode == 1)
                {
                if(vlocy > FILELEN)  /* Outside the last line? Just move */
                    {
                    for(count=locx+1;count<COLUMNS;count++)
                        {
                        if(count% tabsize == 0) break;
                        }

                    locx = count;
                    if(locx > COLUMNS) locx-=COLUMNS;
                    MoveCursor(locx,locy);
                    break;
                    }

                if (locx > strlen(lines[vlocy]))   /* Pad chars to us */
                    {
                    padchars(strlen(lines[vlocy]), locx, lines[vlocy]);
                    }

                tmp=locx;

                for(count=locx+1;count<COLUMNS;count++)
                    {
                    if(count % tabsize == 0) break;
                    }

                locx = count;

                if(locx >= COLUMNS) locx = 0;

                tmp1 = locx - tmp;

                if(insertchars(tmp1,tmp,' ',lines[vlocy]) == 1)
                    {
                    RedrawLine(locx,locy,vlocy);
                    }
                else
                    {
                    MoveCursor(locx,locy);
                    putchar('\a');
                    }
                }
            else

/******************************************/
/*                                        */
/* No insertmode, no blocks, no nuthin    */
/* just move the cursor.                  */
/*                                        */
/******************************************/

                {
                for(count=locx+1;count<COLUMNS;count++)
                    {
                    if(count% tabsize == 0) break;
                    }

                locx = count;

                if(locx >= COLUMNS) locx=0;

                MoveCursor(locx,locy);
                break;
                }
            break;
            }

/******************************************/
/*                                        */
/* The RTAB is very similar to TAB in     */
/* that if there is a marking region set  */
/*                           we want to   */
/* delete the first tabsize spaces.       */
/* from the beginning of each line, else  */
/* just move the cursor.                  */
/*                                        */
/******************************************/

        case RTAB:
            {
            if(MARKING == 0)
                {
                for(count=locx-1;count>=0;count--)
                    {
                    if(count % tabsize == 0) break;
                    }

                locx = count;
                if(locx <0) locx+=COLUMNS;
                }
            else
                {
                for(count=STARTBUF;count<=ENDBUF;count++)
                    {
                    if(lines[STARTBUF] == NULL || lines[ENDBUF] == NULL) break;

/******************************************/
/*                                        */
/* Compute how many spaces are available  */
/* to delete from this line.              */
/*                                        */
/******************************************/

                    for(count1=0;count1<tabsize;count1++)
                        {
                        if(!isspace(lines[count][count1])) break;
                        }

                    tmp=count1;

                    if(tmp == 0) continue;

                    deletechars(tmp,0,lines[count]);
                    if(count >= vlocy-locy &&
                        count <= vlocy-locy+LINES)
                        {
                        RedrawLine(0,count-(vlocy-locy),count);
                        }
                    }
                }

            MoveCursor(locx,locy);
            break;
            }

/******************************************/
/*                                        */
/* For CRLF if we are in insert mode, we  */
/* need to break the line into to, and    */
/* continue it on the next line.  Doing   */
/* whatever scrolling and repaingint is   */
/* necessary.  If we arent in insertmode  */
/* we simple move the cursor to the       */
/* beginning of the next line.            */
/*                                        */
/******************************************/

        case CRLF:
crlf:
            {
            if(insertmode == 1 && MARKING == 0)   /* We in insert! */
                {
                if (vlocy >= MAXLINE) break;      /* Outside of range */

                if (FILELEN+1 > MAXLINE) break;   /* Cant split out of range */

                if (vlocy > FILELEN)              /* We need more lines! */
                    {
                    padlines(vlocy, lines);
                    }

/******************************************/
/*                                        */
/* Snip the part we are going to put on   */
/* the next line, and copy it to the      */
/* buffer.                                */
/*                                        */
/******************************************/

                for(count=locx;lines[vlocy][count];count++)
                    {
                    linebuf[count-locx] = lines[vlocy][count];
                    }

                linebuf[count-locx] = '\0';

/******************************************/
/*                                        */
/*  Remove the cut part from the string   */
/*                                        */
/******************************************/

                for(count=locx;count<CALLOCSIZE;count++)
                    {
                    lines[vlocy][count] = '\0';
                    }

/******************************************/
/*                                        */
/* Clear the cut part from the screan,    */
/* position the cursor to the next line,  */
/* insert a line, and copy the cut part   */
/* to this new line.                      */
/*                                        */
/******************************************/

                if(locx < COLUMNS) CleartoEOLN();
                vlocy++;
                locy++;
                insertlines(lines, vlocy, 1);
                strcpy(lines[vlocy],linebuf);

/******************************************/
/*                                        */
/* If we need to scroll do it here, else  */
/* just do the required IL/DL functions   */
/* to open a line here.                   */
/*                                        */
/******************************************/

                if(locy > LINES - 1)
                    {
                    scrolldown(&locx, &locy, &vlocy, lines);
                    }
                else
                    {
                    MoveCursor(0,LINES);
                    DeleteLine();
                    MoveCursor(0,locy);
                    InsertLine();
                    RedrawLine(locx,locy,vlocy);
                    }
                }
            else
                {

/******************************************/
/*                                        */
/* If we are at the line at the very end  */
/* of the file (MAXLINE) we dont want to  */
/* move the cursor to the next line.      */
/* If we arent, move the cursor to the    */
/* next line.  If we are marking we need  */
/* to extend the marking region at this   */
/* point too.                             */
/*                                        */
/******************************************/

                if(MARKING && vlocy >= FILELEN) break;

                if(vlocy >= MAXLINE) break;

                vlocy++;
                locy++;
                }

            locx=lcol;
            cret=MoveCursor(locx,locy);

            if(MARKING)
                {
                if(vlocy > STARTBUF)
                    {
                    ENDBUF=vlocy;
                    if(locy <= LINES) RedrawLine(locx,locy,vlocy);
                    }
                }
            break;
            }

/******************************************/
/*                                        */
/* If the use has pressed backspace, and  */
/* we are in insertmode we need to move   */
/* the cursor to the left, blasting the   */
/* character to the left of the cursor    */
/* first.  If we arent in insertmode, we  */
/* drop through to the LEFT case.         */
/*                                        */
/******************************************/

        case BACKSPACE:
            {
            if(insertmode)
                {
                locx--;
                if(locx < 0)
                    {
                    locx=COLUMNS - 1;
                    MoveCursor(locx,locy);
                    }
                deletechars(1, locx, lines[vlocy]);
                MoveCursor(locx,locy);
                RedrawFromX(locx, locy, lines[vlocy]);
                putchar(' ');
                MoveCursor(locx,locy);
                break;
                }
            }

/******************************************/
/*                                        */
/* They hit left, simple enough just move */
/* the cursor left!                       */
/*                                        */
/******************************************/

        case LEFT:
            {
            putchar('\b');
            locx--;

            if(locx < 0)
                {
                cret=MoveCursor(locx,locy);
                }

            break;
            }

/******************************************/
/*                                        */
/* They hit right, move the cursor to the */
/* right.                                 */
/*                                        */
/******************************************/

        case RIGHT:
            {
            if(lines[vlocy] != NULL)
                {
                if(lines[vlocy][locx] != '\0')
                    {
                    putchar(lines[vlocy][locx]);
                    locx++;
                    }
                else
                    {
                    putchar(' ');
                    locx++;
                    }
                }
            else
                {
                locx++;
                cret=MoveCursor(locx,locy);
                break;
                }

            if(locx > COLUMNS-1)
                {
                cret=MoveCursor(locx,locy);
                }

            break;
            }

/******************************************/
/*                                        */
/* DEL wipes out the current character    */
/* under the cursor, pulling the line     */
/* back one.                              */
/*                                        */
/******************************************/

        case DEL:
            {
            deletechars(1, locx, lines[vlocy]);
            RedrawFromX(locx, locy, lines[vlocy]);
            putchar(' ');
            MoveCursor(locx,locy);
            break;
            }

/******************************************/
/*                                        */
/* F4 is used to quit without saving, we  */
/* make sure the user really wants to by  */
/* requiring a capital Y to quit.         */
/*                                        */
/******************************************/

        case FKEY4:

            getdata("Exit without save, ARE YOU SURE? (CONFIRM WITH CAPITAL "
                    "Y)","N", ctmp,locx,locy,vlocy,lines,1);

            if (ctmp[0] != 'Y') break;

            Raw(0);

            ClearScreen();

            printf("CE %s Copyright (C) 1991-92, Charles R. "
                   "Henrich, All Rights Reserved.\n\r\n\r",VERSION);

            clean_exit(0);
            break;

#ifndef NOSHELL

/******************************************/
/*                                        */
/* For both F6 and F7 we clear the screen */
/* save the current text to disk, and     */
/* then run the shell command on it.      */
/* We have to leave raw mode before       */
/* executing the shell script however.    */
/*                                        */
/******************************************/

        case FKEY6:

            savefile(fname,wordwrap,0,nobaks);

            Raw(0);

            ClearScreen();

/******************************************/
/*                                        */
/* When we parse the filename we want to  */
/* names to emerge, the first being the   */
/* full filename, and the second being    */
/* the filename without the last .<stuff> */
/* ctmp will hold the stripped filename,  */
/* if we cant strip off a .<stuff> we     */
/* call the second name shell#            */
/*                                        */
/******************************************/

            strcpy(ctmp2,fname);

            ptmp=(char *)rindex(ctmp2,'.');

            if(ptmp != NULL) *ptmp='\0';
            else
            strcpy(ctmp2,"shell1");

            sprintf(ctmp,SHELL1,fname, ctmp2);
            printf("Executing %s\n\n",ctmp);

            system(ctmp);

            printf("\nPress RETURN to Continue");

            getkey(&cha,&func);

            Raw(1);

            RedrawScreen(vlocy-locy,lines,locx,locy);
            break;

/******************************************/
/*                                        */
/* Shell escape #2                        */
/*                                        */
/******************************************/

        case FKEY7:

            savefile(fname,wordwrap,0,nobaks);

            Raw(0);

            ClearScreen();

            strcpy(ctmp2,fname);
            ptmp=(char *)rindex(ctmp2,'.');

            if(ptmp != NULL) *ptmp='\0';
            else
            strcpy(ctmp2,"shell2");

            sprintf(ctmp,SHELL2,fname, ctmp2);
            printf("Executing %s\n\n",ctmp);
            system(ctmp);

            printf("\nPress RETURN to Continue");

            getkey(&cha,&func);

            Raw(1);

            RedrawScreen(vlocy-locy,lines,locx,locy);
            break;

#endif  /* NO SHELL */

/******************************************/
/*                                        */
/* Save and exit, allowing the user to    */
/* change the current filename.           */
/*                                        */
/******************************************/

#ifndef NOALTNAME

        case FKEY3:

            getdata("Save as",fname,ctmp,locx,locy,vlocy,lines,0);
            savefile(ctmp,wordwrap,1,nobaks);
            break;

#endif

/******************************************/
/*                                        */
/* Save and dont exit.  Straight forward  */
/*                                        */
/******************************************/

        case FKEY2:

            strcpy(STATUSMSG,"Saving....One Moment Please...");
            drawstatus(locx,locy);
            savefile(fname, wordwrap, 0,nobaks);
            strcpy(STATUSMSG,DEFAULT_STATUSMSG);
            drawstatus(locx,locy);
            break;

/******************************************/
/*                                        */
/* F9 is a toggle to turn marking on/off  */
/* In either case we need to set the      */
/* status mode to either marking or not   */
/* If we are just starting marking we     */
/* set the Marking mode to on, as well as */
/* setting the begin and ending buffer    */
/* marks to the current line.             */
/*                                        */
/******************************************/

        case FKEY9:

            if(MARKING)
                {
                strcpy(STATUSMSG,DEFAULT_STATUSMSG);

                if(insertmode) strcpy(STATUSMODE,"INSERT");
                else strcpy(STATUSMODE,"OVERWRITE");

                drawstatus(locx,locy);
                MARKING=0;

                break;
                }

            strcpy(STATUSMSG,"ESC 9 To stop marking");
            strcpy(STATUSMODE,"MARKING");
            drawstatus(locx,locy);
            MARKING=1;

/******************************************/
/*                                        */
/* Some optimizations, if there was no    */
/* previous marking region set, just      */
/* set the marking region to the current  */
/* line, and repaint it, else repaint the */
/* entire screen.                         */
/*                                        */
/******************************************/


            if(STARTBUF != -1 || ENDBUF != -1)
                {
                STARTBUF = ENDBUF = vlocy;
                RedrawScreen(vlocy-locy,lines,locx,locy);
                }
            else
                {
                STARTBUF = ENDBUF = vlocy;
                RedrawLine(locx,locy,vlocy);
                }

            break;

/******************************************/
/*                                        */
/* Clear the marking region, we only want */
/* to repaint the screen if the region is */
/* visible!  We reset the status mode     */
/* here as well.  Id we dont repaint the  */
/* screen we have to update the status    */
/* line independantly.                    */
/*                                        */
/******************************************/

        case FKEY10:

            strcpy(STATUSMSG,DEFAULT_STATUSMSG);

            if(insertmode) strcpy(STATUSMODE,"INSERT");
            else strcpy(STATUSMODE,"OVERWRITE");

            MARKING=0;

            if((STARTBUF >= vlocy-locy && STARTBUF <= vlocy-locy+LINES) ||
               (ENDBUF >= vlocy-locy && ENDBUF <= vlocy-locy+LINES))
                {
                STARTBUF=ENDBUF=-1;
                RedrawScreen(vlocy-locy,lines,locx,locy);
                }
            else
                {
                STARTBUF=ENDBUF=-1;
                drawstatus(locx,locy);
                }

            break;

/******************************************/
/*                                        */
/* Insert marked text at current line.    */
/*                                        */
/******************************************/

        case FKEY11:

/******************************************/
/*                                        */
/* If the marking region isnt set, we     */
/* abort the operation, if the end or     */
/* beginning of the marking region is     */
/* beyond the end of the file, we make it */
/* the end of this file.  If the          */
/* beginning or end of the region is not  */
/* allocated we abort the operation.      */
/*                                        */
/******************************************/

            if(STARTBUF == -1 || ENDBUF == -1) break;
            if(ENDBUF > FILELEN) ENDBUF = FILELEN;
            if(STARTBUF > FILELEN) STARTBUF = FILELEN;
            if(lines[STARTBUF] == NULL|| lines[ENDBUF] == NULL) break;

/******************************************/
/*                                        */
/* We free anything that was left from a  */
/* previous buffer usage, then copy the   */
/* marked region into the buffer.         */
/*                                        */
/******************************************/

            for(count=0;count<=BUFLEN;count++)
                {
                if(buffer[count] != NULL) free(buffer[count]);
                buffer[count]=0;
                }

            tmp=FILELEN;
            FILELEN=0;
            BUFLEN=ENDBUF-STARTBUF;

            for(count=0;count<=BUFLEN;count++)
                {
                padlines(count, buffer);
                strcpy(buffer[count],lines[count+STARTBUF]);
                }

            FILELEN=tmp;

/******************************************/
/*                                        */
/* We now try and insert BUFLEN number of */
/* blank lines into the real text.  If    */
/* the insertion was successful we copy   */
/* the buffer into the newly inserted     */
/* lines.  Else we beep and abort.  We    */
/* repaint the screen in either case just */
/* to be "sure" about the current state   */
/* of things.                             */
/*                                        */
/******************************************/

            if (vlocy > FILELEN)
                {
                padlines(vlocy, lines);
                }

            if(insertlines(lines,vlocy,BUFLEN+1) == 1)
                {
                for(count=0;count<=BUFLEN;count++)
                    {
                    strcpy(lines[count+vlocy],buffer[count]);
                    }
                }
            else
                {
                putchar('\a');
                }

            RedrawScreen(vlocy-locy, lines,locx,locy);
            break;

/******************************************/
/*                                        */
/* Move the marked region to the current  */
/* line.                                  */
/*                                        */
/******************************************/

        case FKEY12:

/******************************************/
/*                                        */
/* If the marking region isnt set, we     */
/* abort the operation, if the end or     */
/* beginning of the marking region is     */
/* beyond the end of the file, we make it */
/* the end of this file.  If the          */
/* beginning or end of the region is not  */
/* allocated we abort the operation.      */
/*                                        */
/******************************************/

            if(STARTBUF == -1 || ENDBUF == -1) break;
            if(ENDBUF > FILELEN) ENDBUF = FILELEN;
            if(STARTBUF > FILELEN) STARTBUF = FILELEN;
            if(lines[STARTBUF] == NULL|| lines[ENDBUF] == NULL) break;

/******************************************/
/*                                        */
/* We free anything that was left from a  */
/* previous buffer usage, then copy the   */
/* marked region into the buffer.         */
/*                                        */
/******************************************/

            for(count=0;count<=BUFLEN;count++)
                {
                if(buffer[count] != NULL) free(buffer[count]);
                buffer[count]=0;
                }

            tmp=FILELEN;
            FILELEN=0;
            BUFLEN=ENDBUF-STARTBUF;

            for(count=0;count<=BUFLEN;count++)
                {
                padlines(count, buffer);
                strcpy(buffer[count],lines[count+STARTBUF]);
                }

            FILELEN=tmp;

/******************************************/
/*                                        */
/* Some really ugly stuff here so we can  */
/* keep the current line position         */
/* throughout the deletion of the old     */
/* block of text.                         */
/*                                        */
/******************************************/

            {
            int start_tmp, end_tmp;

            start_tmp = STARTBUF;
            end_tmp   = ENDBUF;

/******************************************/
/*                                        */
/* If we cant delete the lines, abort     */
/* deletelines will change the beginning  */
/* and ending position of START and END   */
/* so we have to hold them in a temp      */
/* place when we do the delete.           */
/*                                        */
/******************************************/

            if(deletelines(lines, STARTBUF, BUFLEN+1) != 1)
                {
                break;
                }

/******************************************/
/*                                        */
/* Restore the old values of start and    */
/* end.                                   */
/*                                        */
/******************************************/

            STARTBUF = start_tmp;
            ENDBUF   = end_tmp;

/******************************************/
/*                                        */
/* Save the current line number.          */
/*                                        */
/******************************************/

            tmp=vlocy;

/******************************************/
/*                                        */
/* If the current line is outside the     */
/* end of the marked region, move the     */
/* current line back the correct number   */
/* of lines. (I.e. we just erased that    */
/* number of lines.)                      */
/*                                        */
/******************************************/

            if (vlocy > ENDBUF)
                {
                vlocy-=(BUFLEN+1);
                }
            else

/******************************************/
/*                                        */
/* If the current line is somewhere       */
/* inside the old marked region, make the */
/* current line the beginning of the      */
/* region.                                */
/*                                        */
/******************************************/

                {
                if (vlocy >=STARTBUF && vlocy<= ENDBUF)
                    {
                    vlocy=STARTBUF;
                    }
                }

/******************************************/
/*                                        */
/* If the currentline is outside of the   */
/* existing text, add the correct number  */
/* of lines to make it part of the text.  */
/*                                        */
/******************************************/

                if (vlocy > FILELEN)
                    {
                    padlines(vlocy, lines);
                    }

            tmp1=vlocy;

/******************************************/
/*                                        */
/* Save the values of the start and end   */
/* again.  insertlines messes with those  */
/* values...                              */
/*                                        */
/******************************************/

            start_tmp = STARTBUF;
            end_tmp   = ENDBUF;

            insertlines(lines,vlocy,BUFLEN+1);

            STARTBUF = start_tmp;
            ENDBUF   = end_tmp;
            }

/******************************************/
/*                                        */
/* Copy the buffer into the newly         */
/* inserted lines, and turn marking off,  */
/* clearing the marking region as well.   */
/* we need to update the status line here */
/* to.                                    */
/*                                        */
/******************************************/

            for(count=0;count<=BUFLEN;count++)
                {
                strcpy(lines[count+vlocy],buffer[count]);
                }

            STARTBUF = ENDBUF = -1;
            MARKING=0;

            if(insertmode) strcpy(STATUSMODE,"INSERT");
            else strcpy(STATUSMODE,"OVERWRITE");

/******************************************/
/*                                        */
/* reload the current line from what it   */
/* was when we started this whole mess,   */
/* and do a gotoline to get us to the     */
/* proper place in a nice way.            */
/*                                        */
/******************************************/

            vlocy=tmp;

            ret=gotoline(tmp1+1, &locy, &locx, &vlocy, lines);

            if(ret == 1)
                {
                RedrawScreen(vlocy-locy,lines,locx,locy);
                }
            else
                {
                drawstatus(locx,locy);
                }
            break;

/******************************************/
/*                                        */
/* Save and exit, clean and simple.       */
/*                                        */
/******************************************/

        case FKEY1:

            savefile(fname, wordwrap, 1,nobaks);
            break;

/****************************************************************************/
/*                                                                          */
/* The CONTROL case here does a switch as well.  All Control commands are   */
/* parsed here.  If we made it to this point, the user hasnt pressed any    */
/* "special" keys such as F1 through F12.                                   */
/*                                                                          */
/****************************************************************************/

        case CONTROL:
            {
            switch(cha)
                {

/******************************************/
/*                                        */
/* Do the same exact thing as CRLF for    */
/* Control J or M (we should never        */
/* execute this, but just to be safe.. )  */
/*                                        */
/* Read the CRLF comments to understand   */
/* this code!                             */
/*                                        */
/******************************************/

                case 'J':
                case 'M':
                    func = CRLF;
                    goto crlf;
                    break;

/******************************************/
/*                                        */
/* Control H should never be seen either  */
/* but just in case, just move the cursor */
/* left.                                  */
/*                                        */
/******************************************/

                case 'H':
                    locx--;
                    cret=MoveCursor(locx,locy);
                    break;

/******************************************/
/*                                        */
/* Page forward.  We compute the          */
/* beginning of the next page of screen   */
/* by taking the current line, subtractin */
/* the current cursor line from it, and   */
/* then adding the number of lines on     */
/* the screen to it.  If we are marking   */
/* Here, we need to extend the region to  */
/* include everything up until the new    */
/* line.  For various reasons we also     */
/* reset the statusmsg here.  We also     */
/* beep if we are drawing the last screen */
/*                                        */
/******************************************/

                case 'F':

                    if(vlocy-locy+LINES == FILELEN)
                        {
                        putchar('\a');
                        break;
                        }

                    vlocy=vlocy+LINES-locy;

                    if(vlocy > FILELEN || vlocy+LINES > FILELEN)
                        {
                        vlocy=FILELEN-LINES;
                        if(vlocy < 0) vlocy=0;
                        putchar('\a');
                        }

                    if(MARKING)
                        {
                        ENDBUF=vlocy+locy;
                        RedrawLine(locx,locy,vlocy);
                        }

                    strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                    RedrawScreen(vlocy,lines,locx,locy);
                    vlocy=vlocy+locy;
                    break;

/******************************************/
/*                                        */
/* Repaint the current screen, a simple   */
/* call to RedrawScreen.                  */
/*                                        */
/******************************************/

                case 'L':

                    ret=SetScroll(0,LINES);
                    strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                    RedrawScreen(vlocy-locy,lines,locx,locy);
                    break;

/******************************************/
/*                                        */
/* Lets show the user some vital stats    */
/* about the current file here, displays  */
/* it in the statusline.  We are extra    */
/* careful here to never exceed 45 chars  */
/* so when we copy to the statusmsg we    */
/* dont dump core.  Most parameters are   */
/* stored internally starting with 0, so  */
/* we add one to please the user.         */
/*                                        */
/******************************************/

                case 'V':
                    sprintf(ctmp,"%d%d",vlocy+1, FILELEN+1);
                    tmp=strlen(fname)+strlen(ctmp) + 26;

                    if(tmp > 45)
                        {
                        if(tmp - strlen(fname) < 45)
                            {
                            sprintf(STATUSMSG, "Editing line %d out of %d",
                                vlocy+1, FILELEN+1);
                            drawstatus(locx,locy);
                            break;
                            }

                        sprintf(STATUSMSG, "Not enough space to show "
                            "information");
                        drawstatus(locx,locy);
                        break;
                        }
                    else
                        {
                        sprintf(STATUSMSG, "Editing line %d out of %d in %s",
                                vlocy+1,FILELEN+1,fname);
                        drawstatus(locx,locy);
                        break;
                        }
                    break;

/******************************************/
/*                                        */
/* Page backwards, same computation as    */
/* paging forward, except we subtract     */
/* LINES instead of adding.  We need to   */
/* extend the marking region is we are    */
/* marking, as well as beep on the first  */
/* page.                                  */
/*                                        */
/******************************************/

                case 'B':

                    if(vlocy-locy == 0)
                        {
                        putchar('\a');
                        break;
                        }

                    vlocy=vlocy-locy-LINES;

                    if (vlocy < 0)
                        {
                        vlocy=0;
                        putchar('\a');
                        }

                    if(MARKING)
                        {
                        ENDBUF=vlocy+locy;
                        RedrawLine(locx,locy,vlocy);
                        }

                    strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                    RedrawScreen(vlocy,lines,locx,locy);
                    vlocy=vlocy+locy;
                    break;

/******************************************/
/*                                        */
/* Delete a line!  First we need to make  */
/* sure that we are on a deleteable line  */
/* (i.e. inside filelen).  If we are      */
/* we copy the line into the deletebuffer */
/* so the user can recover it with ^A.    */
/* If the delete is valid we do the       */
/* delete on the screen.                  */
/*                                        */
/******************************************/

                case 'D':

                    if(vlocy > MAXLINE) break;
                    if(vlocy <= FILELEN) strcpy(deletebuf,lines[vlocy]);

                    if(deletelines(lines, vlocy, 1) == 1)
                        {
                        DeleteLine();
                        MoveCursor(0,LINES);
                        InsertLine();

                        if (vlocy-locy+LINES <= FILELEN)
                            {
                            RedrawLine(0,LINES, vlocy-locy+LINES);
                            }

                        MoveCursor(locx,locy);
                        }
                    break;

/******************************************/
/*                                        */
/* Joint the next line with the current   */
/* line.                                  */
/* If the next line doesnt exist break.   */
/*                                        */
/* tmp is the right edge of the cur line  */
/*                                        */
/* tmp1 is the number of the first non-   */
/*      space on the next line.           */
/*                                        */
/* tmp2 is the right edge of the next     */
/*      line                              */
/*                                        */
/* ret contains the rightmost edge of     */
/*     the current line.                  */
/*                                        */
/******************************************/

                case 'U':

                    if(lines[vlocy+1] == 0)
                        {
                        putchar ('\a');
                        break;
                        }

                    strip(lines,vlocy,wordwrap);
                    strip(lines,vlocy+1,wordwrap);

                    ret=strlen(lines[vlocy]);

                    tmp=COLUMNS-ret-1;

                    if(tmp < 1)
                        {
                        locy++;
                        vlocy++;
                        cret=MoveCursor(locx,locy);
                        break;
                        }

                    for(count=0;count<COLUMNS;count++)
                        {
                        if(lines[vlocy+1][count] != ' ') break;
                        }

                    tmp1 = count;

                    for(count=strlen(lines[vlocy+1]);count>=0;count--)
                        {
                        if(lines[vlocy+1][count] != ' ') break;
                        }

                    tmp2 = strlen(lines[vlocy+1]);

/******************************************/
/*                                        */
/* If the entire next line will fit on    */
/* the current line, add it to this line  */
/* and then blast the next line.          */
/*                                        */
/******************************************/

                    if((tmp2 - tmp1) < tmp)
                        {
                        strcat(lines[vlocy]," ");
                        strcat(lines[vlocy],&(lines[vlocy+1][tmp1]));

                        if(deletelines(lines, vlocy+1, 1) == 1)
                            {
                            MoveCursor(locx,locy+1);
                            DeleteLine();
                            MoveCursor(0,LINES);
                            InsertLine();

/******************************************/
/*                                        */
/* Only repaint the last line on the      */
/* screen if it exists.                   */
/*                                        */
/******************************************/

                            if (vlocy-locy+LINES <= FILELEN)
                                {
                                RedrawLine(0,LINES, vlocy-locy+LINES);
                                }

/******************************************/
/*                                        */
/* If we didnt just repaint this line,    */
/* redraw it starting from where the      */
/* next line was tacked on.               */
/*                                        */
/******************************************/

                            if(locy < LINES)
                                {
                                MoveCursor(ret,locy);
                                RedrawFromX(ret, locy, lines[vlocy]);
                                if(ret < COLUMNS) CleartoEOLN();
                                }

                            MoveCursor(locx,locy);
                            }

                        break;
                        }

/******************************************/
/*                                        */
/* Okay, if we are here we can only take  */
/* part of the next line, not all of it.  */
/* In this case we want to skip all       */
/* leading white space, and copy on a     */
/* word by word basis.                    */
/*                                        */
/* Start subtracting words from the next  */
/* line, after every subtraction, see if  */
/* what we have left will fit on this     */
/* line, if so, we do a strncat of the    */
/* next line, and then delete the same    */
/* number of characters we copied from    */
/* it.                                    */
/*                                        */
/******************************************/

                    for(count=strlen(lines[vlocy+1])-1;count>=0;count--)
                        {
                        if(lines[vlocy+1][count] != ' ') continue;

                        if((count-tmp1) < tmp)
                            {
                            strcat(lines[vlocy]," ");
                            strncat(lines[vlocy],&(lines[vlocy+1][tmp1]),
                                (count-tmp1));
                            deletechars(count+1, 0, lines[vlocy+1]);
                            tmp = -1;
                            break;
                            }
                        }

/******************************************/
/*                                        */
/* If tmp != -1 we didnt copy anything..  */
/*                                        */
/******************************************/

                    if(tmp != -1)
                        {
                        locy++;
                        vlocy++;
                        cret=MoveCursor(locx,locy);
                        break;
                        }

/******************************************/
/*                                        */
/* If the "next" line is visible, redraw  */
/* it.                                    */
/*                                        */
/******************************************/

                    if(locy < LINES) RedrawLine(locx,locy+1,vlocy+1);

/******************************************/
/*                                        */
/* Redraw the end of this line, where we  */
/* appended the text to it.               */
/*                                        */
/******************************************/

                    MoveCursor(ret,locy);
                    RedrawFromX(ret, locy, lines[vlocy]);
                    if(ret < COLUMNS) CleartoEOLN();
                    MoveCursor(locx,locy);
                    break;

/******************************************/
/*                                        */
/* We want to insert a blank line here.   */
/* If the current line >= the largest     */
/* possible line we abort, if this line   */
/* is > the number of lines in the file   */
/* we pad it with blanks to get here.     */
/* Then its just a simple call to         */
/* insertlines and we are done.           */
/*                                        */
/******************************************/

                case 'P':

                    if (vlocy >= MAXLINE) break;

                    if (vlocy > FILELEN)
                        {
                        padlines(vlocy, lines);
                        }

                    if(insertlines(lines,vlocy,1) == -1) break;

                    MoveCursor(0,LINES);
                    DeleteLine();
                    MoveCursor(0,locy);
                    InsertLine();
                    MoveCursor(locx,locy);
                    break;

/******************************************/
/*                                        */
/* Insert many lines here, we check to    */
/* make sure we wont insert more lines    */
/* than allowed, and if we arent, just    */
/* a simple insertlines call again.       */
/*                                        */
/******************************************/

                case 'K':

                    getdata("Insert how many lines","5",
                            ctmp,locx,locy,vlocy,lines,1);
                    sscanf(ctmp,"%d",&tmp1);

                    if( (vlocy+tmp1) > MAXLINES) break;

                    if (vlocy > FILELEN)
                        {
                        padlines(vlocy, lines);
                        }

                    insertlines(lines,vlocy,tmp1);
                    RedrawScreen(vlocy-locy,lines,locx,locy);
                    break;

/******************************************/
/*                                        */
/* Toggle between insertmode and          */
/* overwrite mode (which is far superior) */
/* Simply copy the appropriate message    */
/* to statusmode and redraw the status    */
/* line, and set the insertmode variable. */
/*                                        */
/******************************************/

                case 'T':
                    if(insertmode)
                        {
                        strcpy(STATUSMODE,"OVERWRITE");
                        insertmode=0;
                        drawstatus(locx,locy);
                        }
                    else
                        {
                        strcpy(STATUSMODE,"INSERT");
                        insertmode=1;
                        drawstatus(locx,locy);
                        }

                    break;

/******************************************/
/*                                        */
/* Same as DEL                            */
/*                                        */
/******************************************/

                case 'W':

                    deletechars(1, locx, lines[vlocy]);
                    RedrawFromX(locx, locy, lines[vlocy]);
                    putchar(' ');
                    MoveCursor(locx,locy);
                    break;

/******************************************/
/*                                        */
/* Move the cursor to the beginning of    */
/* the line, if this line doesnt exist    */
/* move the cursor to column 0;           */
/*                                        */
/******************************************/

                case 'Z':

                    if(MARKING == 0)
                        {
                        if(vlocy > FILELEN)
                            {
                            locx=0;
                            MoveCursor(locx,locy);
                            break;
                            }

/******************************************/
/*                                        */
/* Find the first nonspace character in   */
/* the line.                              */
/*                                        */
/******************************************/

                        for(count=0;lines[vlocy][count];count++)
                            {
                            if (isspace(lines[vlocy][count])) continue;
                            break;
                            }

                        locx=count;
                        MoveCursor(locx,locy);
                        }
                    else
                        {
                        if(ENDBUF > FILELEN) ENDBUF = FILELEN;

                        if(lines[STARTBUF] == NULL || lines[ENDBUF] == NULL)
                            {
                            break;
                            }

                        for(count=STARTBUF;count<=ENDBUF;count++)
                            {
                            for(count1=0;lines[count][count1];count1++)
                                {
                                if(!isspace(lines[count][count1])) break;
                                }

                            tmp=count1;

                            if(tmp == 0) continue;

                            deletechars(tmp,0,lines[count]);

                            if(count >= vlocy-locy &&
                                count <= vlocy-locy+LINES)
                                {
                                RedrawLine(0,count-(vlocy-locy),count);
                                }
                            }
                        MoveCursor(locx,locy);
                        }
                    break;

/******************************************/
/*                                        */
/* Move the cursor to the end of the      */
/* current line.                          */
/*                                        */
/******************************************/

                case 'C':

                    if(MARKING == 0)
                        {
                        if(vlocy > FILELEN) break;

/******************************************/
/*                                        */
/* Find the first nonspace character from */
/* the right edge of the line.  Add one   */
/* to it so the cursor is sitting next    */
/* to the first nonspace character.       */
/*                                        */
/******************************************/

                        for(count=strlen(lines[vlocy])-1;count>=0;count--)
                            {
                            if (isspace(lines[vlocy][count])) continue;
                            break;
                            }

                        locx=count+1;
                        if(locx > COLUMNS-1) locx=COLUMNS-1;
                        MoveCursor(locx,locy);
                        }
                    else
                        {
                        if(ENDBUF > FILELEN) ENDBUF = FILELEN;

                        if(lines[STARTBUF] == NULL || lines[ENDBUF] == NULL)
                            {
                            break;
                            }

                        for(count=STARTBUF;count<=ENDBUF;count++)
                            {
                            strip(lines,count,wordwrap);

                            tmp = COLUMNS - strlen(lines[count]) - 1;

                            insertchars(tmp,0,' ',lines[count]);

                            if(count >= vlocy-locy && count <=
                                vlocy-locy+LINES)
                                {
                                RedrawLine(0,count-(vlocy-locy),count);
                                }
                            }
                        MoveCursor(locx,locy);
                        break;
                        }
                    break;

/******************************************/
/*                                        */
/* Move the cursor to the center of this  */
/* line.                                  */
/*                                        */
/******************************************/

                case 'X':

                    if(MARKING == 0)
                        {
                        if(vlocy > FILELEN) break;

                        for(count=0;lines[vlocy][count];count++)
                            {
                            if(isspace(lines[vlocy][count])) continue;
                            break;
                            }

/******************************************/
/*                                        */
/* Find the right edge....                */
/*                                        */
/******************************************/

                        for(tmp=strlen(lines[vlocy])-1;tmp>=0;tmp--)
                            {
                            if(isspace(lines[vlocy][tmp])) continue;
                            break;
                            }

/******************************************/
/*                                        */
/* Divide those by two and move the       */
/* cursor.                                */
/*                                        */
/******************************************/

                        locx = ((tmp-count)/2)+count;
                        MoveCursor(locx,locy);
                        }
                    else
                        {
                        if(ENDBUF > FILELEN) ENDBUF = FILELEN;

                        if(lines[STARTBUF] == NULL || lines[ENDBUF] == NULL)
                            {
                            break;
                            }

                        for(count=STARTBUF;count<=ENDBUF;count++)
                            {
                            strip(lines,count,wordwrap);

/******************************************/
/*                                        */
/* tmp = The right edge of the string     */
/* count1 = the left edge of the string   */
/* tmp1 = the length of the string;       */
/* tmp2 = How many spaces belong left of  */
/*        the string.                     */
/*                                        */
/******************************************/

                            tmp = strlen(lines[count]);

                            if(tmp == 0) continue;

                            for(count1=0;lines[count][count1];count1++)
                                {
                                if(!isspace(lines[count][count1])) break;
                                }

                            tmp1 = tmp - count1;

                            tmp2 = (COLUMNS - tmp1) / 2;

                            if(count1 == tmp2) continue;

                            if(count1 < tmp2)
                                {
                                insertchars(tmp2-count1,0,' ',lines[count]);
                                }
                            else
                                {
                                deletechars(count1-tmp2,0,lines[count]);
                                }

                            if(count >= vlocy-locy &&
                               count <= vlocy-locy+LINES)
                                {
                                RedrawLine(0,count-(vlocy-locy),count);
                                }
                            }
                        MoveCursor(locx,locy);
                        }

                    break;

/******************************************/
/*                                        */
/* Almost a CONTROL P operation, except   */
/* that when we get a valid insert we     */
/* copy the contents of deletebuf to the  */
/* newly inserted line.                   */
/*                                        */
/******************************************/

                case 'A':

                    if (vlocy >= MAXLINE) break;

                    if (vlocy > FILELEN)
                        {
                        padlines(vlocy, lines);
                        }

                    if(insertlines(lines, vlocy, 1)==-1) break;

                    MoveCursor(0,LINES);
                    DeleteLine();
                    MoveCursor(0,locy);
                    InsertLine();

                    strcpy(lines[vlocy],deletebuf);

                    RedrawLine(locx,locy,vlocy);

                    break;

/******************************************/
/*                                        */
/* Set the left column and tabsizing vars */
/* We use the getdata function here for   */
/* that, very straight forward, with a    */
/* few checks.                            */
/*                                        */
/******************************************/

                case 'O':

                    getdata("Enter new left column",itoa(locx+1),
                            ctmp,locx,locy,vlocy,lines,1);

                    tmp=sscanf(ctmp,"%d",&tmp1);

                    if(tmp1 > 79 || tmp1 == 0) tmp1=1;

                    tmp1--;
                    lcol=tmp1;

                    getdata("Enter number of spaces for TAB",itoa(tabsize),
                            ctmp,locx,locy,vlocy,lines,1)  ;

                    tmp=sscanf(ctmp,"%d",&tmp1);

                    if(tmp1 > 79 || tmp1 == 0) tmp1=1;

                    tabsize=tmp1;
                    break;

/******************************************/
/*                                        */
/* Just like DEL except we prompt for the */
/* number of characters to delete.  If we */
/* have a blocking region however we      */
/* delete the entered number of chars     */
/* from all the lines, and from column 0. */
/* Normally we delete from the current    */
/* column.                                */
/*                                        */
/******************************************/

                case 'E':

                    getdata("Wipe out how many chars",itoa(lastdel),
                            ctmp,locx,locy,vlocy,lines,1);

                    tmp=sscanf(ctmp,"%d",&tmp1);

                    if (tmp1 < 0 )
                        {
                        putchar('\a');
                        break;
                        }
                    else
                        {
                        lastdel = tmp1;

/******************************************/
/*                                        */
/* If STARTBUF or ENDBUF == -1 there is   */
/* no valid marked region.                */
/*                                        */
/******************************************/

                        if(STARTBUF == -1 || ENDBUF == -1)
                            {
                            deletechars(tmp1,locx,lines[vlocy]);
                            RedrawFromX(locx,locy,lines[vlocy]);
                            if(locx < COLUMNS) CleartoEOLN();
                            MoveCursor(locx,locy);
                            }
                        else
                            {
                            for(count=STARTBUF;count<=ENDBUF;count++)
                                {
                                deletechars(tmp1,0,lines[count]);

                                if(count >= vlocy-locy &&
                                    count < vlocy-locy+LINES)
                                    {
                                    RedrawLine(0,count-(vlocy-locy),count);
                                    }

                                }
                            MoveCursor(locx,locy);
                            }
                        }

                    break;

/******************************************/
/*                                        */
/* Prompt for a filename, and try and     */
/* read it in..                           */
/*                                        */
/******************************************/

#ifndef NOIMPORT

                case 'Y':

                    getdata("Read file","",ctmp,locx,locy,vlocy,lines,0);

/******************************************/
/*                                        */
/* If the user didnt enter a name, or we  */
/* cant open the file, or the file is not */
/* a "regular" file, we beep and return.  */
/*                                        */
/******************************************/


                    if(ctmp[0] == '\0') break;

                    ifp=fopen(ctmp,"r");

                    if(ifp == NULL)
                        {
                        printf("\a");
                        strcpy(STATUSMSG,"Cant open file!");
                        drawstatus(locx,locy);
                        break;
                        }

                    ret=stat(ctmp,&buf);

                    if(ret == 0)
                        {
                        if(!(S_ISREG(buf.st_mode)))
                            {
                            fclose(ifp);
                            sprintf(STATUSMSG,"%s is not a file!",ctmp);
                            drawstatus(locx,locy);
                            break;
                            }
                        }

/******************************************/
/*                                        */
/* Free anything in the buffer, and then  */
/* read the file into the buffer.         */
/*                                        */
/******************************************/

                    for(count=0;count<=BUFLEN;count++)
                        {
                        if(buffer[count] != NULL) free(buffer[count]);
                        buffer[count]=NULL;
                        }

                    BUFLEN=-1;

                    if (vlocy > FILELEN)
                        {
                        padlines(vlocy, lines);
                        }

                    strcpy(STATUSMSG,"One moment, reading file...");
                    drawstatus(locx,locy);
                    tmp=FILELEN;
                    FILELEN=0;

/******************************************/
/*                                        */
/* Read the file into the buffer...       */
/*                                        */
/******************************************/

                    for(count=0;count<=MAXLINE;count++)
                        {
                        if(fgets(ctmp, COLUMNS+1, ifp) == NULL) break;
                        padlines(count, buffer);
                        strcpy(buffer[count],ctmp);
                        strip(buffer,count,wordwrap);
                        FILELEN = count;
                        }

                    BUFLEN=FILELEN;
                    FILELEN=tmp;

/******************************************/
/*                                        */
/* Try and insert the appropriate number  */
/* of blank lines in the file if we cant  */
/* set the statusmsg, clear the buffer    */
/* and return.                            */
/*                                        */
/******************************************/

                    if(insertlines(lines,vlocy,BUFLEN+1) != 1)
                        {
                        strcpy(STATUSMSG,
                            "File to large to fit into buffer!\a");
                        drawstatus(locx,locy);

                        for(count=0;count<=BUFLEN;count++)
                            {
                            if(buffer[count] != NULL)free(buffer[count]);
                            buffer[count]=0;
                            }

                        BUFLEN=-1;
                        sleep(3);

                        strcpy(STATUSMSG,DEFAULT_STATUSMSG);

                        drawstatus(locx,locy);
                        break;
                        }

/******************************************/
/*                                        */
/* Okay, we inserted the right number of  */
/* lines, so copy the buffer into the     */
/* text file.  And then free the buffer.  */
/*                                        */
/******************************************/

                    for(count=0;count<=BUFLEN;count++)
                        {
                        strcpy(lines[count+vlocy],buffer[count]);
                        }

                    for(count=0;count<=BUFLEN;count++)
                        {
                        if(buffer[count] != NULL)free(buffer[count]);
                        buffer[count]=0;
                        }

                    BUFLEN=-1;
                    strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                    RedrawScreen(vlocy-locy,lines,locx,locy);
                    break;

#endif  /* NO IMPORT */

/******************************************/
/*                                        */
/* Delete the marked region.  If there is */
/* no marked region, abort.  Make sure    */
/* that the ENDBUF contains a real line,  */
/* if its off the end of the file, make   */
/* it the end of the file.                */
/*                                        */
/******************************************/

                case 'N':

                    if(STARTBUF == -1 || ENDBUF == -1) break;

                    if(ENDBUF > FILELEN) ENDBUF = FILELEN;

                    BUFLEN=ENDBUF-STARTBUF+1;

                    getdata("Delete marked text, ARE YOU SURE? (CONFIRM WITH "
                            "CAPITAL Y)","N", ctmp,locx,locy,vlocy,lines,1);

                    if (ctmp[0] != 'Y') break;

/******************************************/
/*                                        */
/* Try and delete the lines the user      */
/* wants us to.                           */
/*                                        */
/******************************************/

                    if(deletelines(lines, STARTBUF, BUFLEN ) != 1)
                        {
                        break;
                        }

                    tmp=STARTBUF;
                    STARTBUF = ENDBUF = -1;

                    strcpy(STATUSMSG,DEFAULT_STATUSMSG);
                    if(insertmode) strcpy(STATUSMODE,"INSERT");
                    else strcpy(STATUSMODE,"OVERWRITE");

                    MARKING = 0;

                    ret=gotoline(tmp+1, &locy, &locx, &vlocy, lines);

                    if(ret == 1)
                        {
                        RedrawScreen(vlocy-locy,lines,locx,locy);
                        }

                    break;

/******************************************/
/*                                        */
/* Search the text file for a string.     */
/*                                        */
/******************************************/

                case 'R':

                    getdata("Enter Search String",lastsearch,
                            ctmp,locx,locy,vlocy,lines,0);

                    if(ctmp[0] == '\0') break;
                    ret=search(ctmp,&locy,&locx,&vlocy);
                    strcpy(lastsearch,ctmp);
                    break;

/******************************************/
/*                                        */
/* Goto a line...                         */
/*                                        */
/******************************************/

                case 'G':

                    getdata("Goto Line","1",ctmp,locx,locy,vlocy,lines,1);
                    sscanf(ctmp,"%d",&tmp1);
                    gotoline(tmp1, &locy, &locx, &vlocy, lines);
                    break;

/******************************************/
/*                                        */
/* Unknown control key, just toss it.     */
/*                                        */
/******************************************/

                default:

                    fflush(stdin);
                    fflush(stdout);
                    break;
                }

            break;                               /* End of CONTROL case */
            }
        }                                        /* End of SWITCH */

/******************************************/
/*                                        */
/* If ret contains SCROLLUP or SCROLLDOWN */
/* we do the appropriate scrolling here.  */
/*                                        */
/******************************************/

    if(cret==SCROLLUP)
        {
        scrollup(&locx, &locy, &vlocy, lines);
        }

    if(cret==SCROLLDOWN)
        {
        scrolldown(&locx, &locy, &vlocy, lines);
        }

    if(cret==WRAPLEFT) locx=COLUMNS-1;

    if(cret==WRAPRIGHT) locx=0;

/******************************************/
/*                                        */
/* Reset the return value to 0            */
/*                                        */
/******************************************/

    cret=0;
    }
}

/****************************************************************************/
/*                                                                          */
/*  Function: hangup                    Author: Charles Henrich             */
/*                                      Date  : 9-13-92                     */
/*                                                                          */
/*  Purpose: To save the users document when they hang up on the connection */
/*           It is a very straightforward save, and saves into the file     */
/*           cefile.hang                                                    */
/*                                                                          */
/****************************************************************************/

void hangup()
{
FILE *ofp;
int count;

ofp=fopen("cefile.hang","w");

for(count=0;count<=FILELEN;count++)
    {
    if(wordwrap == 0)
        {
        if(strlen(lines[count])>79)
            {
            if(ofp != 0) fprintf(ofp,"%s",lines[count]);
            continue;
            }
        }

    if(ofp != 0) fprintf(ofp,"%s\n", lines[count]);
    }

fclose(ofp);
exit(-5);
}

/****************************************************************************/
/*                                                                          */
/*  Function: winchange                 Author: Charles Henrich             */
/*                                      Date  : 9-13-92                     */
/*                                                                          */
/*  Purpose: This function is called on detection of a window resizing.     */
/*           It makes the appropriate changes to the size variables in CE,  */
/*           and calls a RedrawScreen to update the user, we bitch if the   */
/*           user has resized to less than 80 columns.                      */
/*                                                                          */
/****************************************************************************/

void winchange()
{
ScreenSize(&LINES,&COLUMNS);

LINES--;

if(G_mini > 0) LINES-=MINILINES;

if(locy > LINES)
    {
    vlocy=vlocy-locy;
    locy=0;
    }

/* This editor must have 80 or more columns, if they re-sized the window
   to less, complain LOUDLY */

if (COLUMNS < 80)
    {
    ClearScreen();
    printf("CE Requires 80 columns or more\n\r");
    printf("Please resize your window to this size\n\r");
    MoveCursor(0,0);
    return;
    }

if (COLUMNS > 80) COLUMNS=80;

SetScroll(0,LINES);
RedrawScreen(vlocy-locy,lines,locx,locy);
}

/****************************************************************************/
/*                                                                          */
/*  Function: majerr                    Author: Charles Henrich             */
/*                                      Date  : 9-13-92                     */
/*                                                                          */
/*  Purpose: To cleanly exit on a major error such as seg fault or core     */
/*           dump.  We (being nice guys) attempt to save the current text   */
/*           file into a file called cefile.error.  We also tell the user   */
/*           to mail us about the bug.                                      */
/*                                                                          */
/****************************************************************************/

void majerr()
{
FILE *ofp;
int count;
ofp=fopen("cefile.error","w");

for(count=0;count<=FILELEN;count++)
    {
    if(wordwrap == 0)
        {
        if(strlen(lines[count])>79)
            {
            if(ofp != 0) fprintf(ofp,"%s",lines[count]);
            continue;
            }
        }
    if(ofp != 0) fprintf(ofp,"%s\n", lines[count]);
    }

fclose(ofp);

MoveCursor(0,0);
ClearScreen();
printf("\n\rWarning! Warning!  A Major Error has occurred, CE FAILING\n\r\a");
printf("Your file should be saved in cefile.error\a\n\r\n\r");
printf("Please send a email message immediatly to henrich@crs.cl.msu.edu\n\r");
printf("Describe exactly what you were doing when this error occurred!\n\r\a");
MoveCursor(0,10);
Raw(0);
exit(-200);
}

void dumpfile()
{
strcpy(STATUSMSG,"Paging space low, emergency save...");
drawstatus(locx,locy);
savefile("cefile.hang", wordwrap, 0, 1);
strcpy(STATUSMSG,DEFAULT_STATUSMSG);
drawstatus(locx,locy);
}
