/*****************************************************************************
 *  The Elm Mail System  -  $Revision: 4.1 $   $State: Exp $
 *
 *                      Copyright (c) 1986, 1987 Dave Taylor
 *                      Copyright (c) 1988, 1989, 1990 USENET Community Trust
*****************************************************************************
 * Bug reports, patches, comments, suggestions should be sent to:
 *
 *      Syd Weinstein, Elm Coordinator
 *      elm@DSI.COM                     dsinc!elm
 *
******************************************************************************
 * $Log:        curses.c,v $
 * Revision 4.1  90/04/28  22:42:39  syd
 * checkin of Elm 2.3 as of Release PL0
 *
 *
******************************************************************************
 *
 * Modified for use in CE by Charles Henrich (1991-92)
 *
*****************************************************************************/

/**  This library gives programs the ability to easily access the
     termcap information and write screen oriented and raw input
     programs.  The routines can be called as needed, except that
     to use the cursor / screen routines there must be a call to
     InitScreen() first.  The 'Raw' input routine can be used
     independently, however.

**/

/** NOTE THE ADDITION OF: the #ifndef ELM stuff around routines that
    we don't use.  This is for code size and compile time speed...
**/

#include <stdio.h>
#include <fcntl.h>
#include "curses.h"
#include "ce.h"

#ifdef  BSD
#include <setjmp.h>
#endif

#ifdef TERMIOS
#  include <termios.h>
# if __convex__
#  include <sys/ioctl.h>    /* non-standard, for TIOCGWINSZ */
# endif
#else
# ifdef TERMIO
#  include <termio.h>
# else
#  include <sgtty.h>
# endif
#endif

#include <ctype.h>

#ifdef PTEM
#  include <sys/types.h>
#  include <sys/stream.h>
#  include <sys/ptem.h>
#endif

#ifdef BSD
#undef tolower
#endif

#define TTYIN    0

#ifdef SHORTNAMES
# define _clearinverse    _clrinv
# define _cleartoeoln    _clrtoeoln
# define _cleartoeos    _clr2eos
# define _transmit_off    xmit_off
# define _transmit_on    xmit_on
#endif

#ifdef TERMIOS
struct termios _raw_tty,
           _original_tty;
#define    ttgetattr(fd,where)    tcgetattr((fd),(where))
#define    ttsetattr(fd,where)    tcsetattr((fd),TCSADRAIN,(where))
#else    /*TERMIOS*/
# ifdef TERMIO
struct termio _raw_tty,
              _original_tty;
#define    ttgetattr(fd,where)    ioctl((fd),TCGETA,(where))
#define    ttsetattr(fd,where)    ioctl((fd),TCSETAW,(where))
# else
struct sgttyb _raw_tty,
          _original_tty;
#define    ttgetattr(fd,where)    ioctl((fd),TIOCGETP,(where))
#define    ttsetattr(fd,where)    ioctl((fd),TIOCSETP,(where))
# endif    /*TERMIO*/
#endif    /*TERMIOS*/

static int _inraw = 0;                  /* are we IN rawmode?    */

#define DEFAULT_LINES_ON_TERMINAL    24
#define DEFAULT_COLUMNS_ON_TERMINAL    80

static int _memory_locked = 0;        /* are we IN memlock??   */
static int _line  = -1,            /* initialize to "trash" */
           _col   = -1;

static int _intransmit;            /* are we transmitting keys? */

static
char *_clearscreen, *_moveto, *_up, *_down, *_right, *_left,
     *_setbold, *_clearbold, *_setunderline, *_clearunderline,
     *_sethalfbright, *_clearhalfbright, *_setinverse, *_clearinverse,
     *_cleartoeoln, *_cleartoeos, *_transmit_on, *_transmit_off,
     *_set_memlock, *_clear_memlock, *_insert_line, *_delete_line,
     *_set_scroll,  *_termbegin, *_termend;

static int _lines, _columns, _automargin, _eatnewlineglitch;
int tabspacing;

static char _terminal[1024];              /* Storage for terminal entry */
static char _capabilities[1024];           /* String for cursor motion */

static char *ptr = _capabilities;    /* for buffering         */

int    outchar();            /* char output for tputs */
char  *tgetstr(),                    /* Get termcap capability */
      *tgoto();                /* and the goto stuff    */

InitScreen()
{
    /* Set up all this fun stuff: returns zero if all okay, or;
        -1 indicating no terminal name associated with this shell,
        -2..-n  No termcap for this terminal type known
   */

    int  tgetent(),      /* get termcap entry */
         err;
    char termname[40];
    char *getenv();

    if (getenv("TERM") == NULL) return(-1);

    if (strcpy(termname, getenv("TERM")) == NULL)
        return(-1);

    if ((err = tgetent(_terminal, termname)) != 1)
        return(err-2);

    _line  =  0;        /* where are we right now?? */
    _col   =  0;        /* assume zero, zero...     */

    /* load in all those pesky values */
    _termbegin         = tgetstr("ti", &ptr);
    _termend           = tgetstr("te", &ptr);
    _clearscreen       = tgetstr("cl", &ptr);
    _moveto            = tgetstr("cm", &ptr);
    _up                = tgetstr("up", &ptr);
    _down              = tgetstr("do", &ptr);
    _right             = tgetstr("nd", &ptr);
    _left              = tgetstr("bs", &ptr);
    _setbold           = tgetstr("so", &ptr);
    _clearbold         = tgetstr("se", &ptr);
    _setunderline      = tgetstr("us", &ptr);
    _clearunderline    = tgetstr("ue", &ptr);
    _setinverse        = tgetstr("so", &ptr);
    _clearinverse      = tgetstr("se", &ptr);
    _sethalfbright     = tgetstr("hs", &ptr);
    _clearhalfbright   = tgetstr("he", &ptr);
    _cleartoeoln       = tgetstr("ce", &ptr);
    _cleartoeos        = tgetstr("cd", &ptr);
    _lines                 = tgetnum("li");
    _columns       = tgetnum("co");
    tabspacing       = ((tabspacing=tgetnum("it"))==-1 ? 8 : tabspacing);
    _automargin       = tgetflag("am");
    _eatnewlineglitch   = tgetflag("xn");
    _transmit_on       = tgetstr("ks", &ptr);
    _transmit_off      = tgetstr("ke", &ptr);
    _set_memlock       = tgetstr("ml", &ptr);
    _clear_memlock       = tgetstr("mu", &ptr);
    _delete_line       = tgetstr("dl", &ptr);
    _insert_line       = tgetstr("al", &ptr);
    _set_scroll        = tgetstr("cs", &ptr);

    return(0);
}

char *return_value_of(termcap_label)
char *termcap_label;
{
    /** This will return the string kept by termcap for the
        specified capability. Modified to ensure that if
        tgetstr returns a pointer to a transient address
        that we won't bomb out with a later segmentation
        fault (thanks to Dave@Infopro for this one!)

        Tweaked to remove padding sequences.
     **/

    static char escape_sequence[20];
    register int i=0,j=0;
    char buffer[20];
    char *myptr, *tgetstr();             /* Get termcap capability */

    if (strlen(termcap_label) < 2)
      return(NULL);

    if (termcap_label[0] == 's' && termcap_label[1] == 'o')
      {
      if (_setinverse)
        strcpy(escape_sequence, _setinverse);
      else
        return( (char *) NULL );
      }
    else if (termcap_label[0] == 's' && termcap_label[1] == 'e')
      {
      if (_clearinverse)
        strcpy(escape_sequence, _clearinverse);
      else
        return( (char *) NULL );
      }
    else if ((myptr = tgetstr(termcap_label, &ptr)) == NULL)
      return( (char *) NULL );
    else
      strcpy(escape_sequence, myptr);

    if (chloc(escape_sequence, '$') != -1) {
      while (escape_sequence[i] != '\0') {
        while (escape_sequence[i] != '$' && escape_sequence[i] != '\0')
          buffer[j++] = escape_sequence[i++];
        if (escape_sequence[i] == '$') {
          while (escape_sequence[i] != '>') i++;
          i++;
        }
      }
      buffer[j] = '\0';
      strcpy(escape_sequence, buffer);
    }

    return( (char *) escape_sequence);
}


/****** now into the 'meat' of the routines...the cursor stuff ******/

ScreenSize(lines, columns)
int *lines, *columns;
{
    /** returns the number of lines and columns on the display. **/

#ifdef TIOCGWINSZ
    struct winsize w;

    if (ioctl(1,TIOCGWINSZ,&w) != -1) {
        _lines = w.ws_row;
        _columns = w.ws_col;
    }
#endif

    if (_lines == 0) _lines = DEFAULT_LINES_ON_TERMINAL;
    if (_columns == 0) _columns = DEFAULT_COLUMNS_ON_TERMINAL;

    *lines = _lines - 1;        /* assume index from zero */
    *columns = _columns;
}

SetScroll(start, end)
int start, end;
{
char *termstr;

if(! _set_scroll) return -1;
termstr=tgoto(_set_scroll, end, start);
tputs(termstr, 1, outchar);
return 1;
}

SetXYLocation(x,y)
int x,y;
{
    /* declare where the cursor is on the screen - useful after using
     * a function that moves cursor in predictable fasion but doesn't
     * set the static x and y variables used in this source file -
     * e.g. getpass().
     */

    _line = x;
    _col = y;
}

GetXYLocation(x,y)
int *x,*y;
{
    /* return the current cursor location on the screen */

    *x = _line;
    *y = _col;
}

ClearScreen()
{
    /* clear the screen: returns -1 if not capable */

    _line = 0;    /* clear leaves us at top... */
    _col  = 0;

    if (!_clearscreen)
        return(-1);

    tputs(_clearscreen, 1, outchar);
    fflush(stdout);      /* clear the output buffer */
    return(0);
}

static
CursorUp(n)
int n;
{
    /** move the cursor up 'n' lines **/
    /** Calling function must check that _up is not null before calling **/

    _line = (_line-n > 0? _line - n: 0);    /* up 'n' lines... */

    while (n-- > 0)
        tputs(_up, 1, outchar);

    fflush(stdout);
    return(0);
}


static
CursorDown(n)
int n;
{
    /** move the cursor down 'n' lines **/
    /** Caller must check that _down is not null before calling **/

    _line = (_line+n <= LINES? _line + n: LINES);    /* down 'n' lines... */

    while (n-- > 0)
        tputs(_down, 1, outchar);

    fflush(stdout);
    return(0);
}


static
CursorLeft(n)
int n;
{
    /** move the cursor 'n' characters to the left **/
    /** Caller must check that _left is not null before calling **/

    _col = (_col - n> 0? _col - n: 0);    /* left 'n' chars... */

    while (n-- > 0)
        tputs(_left, 1, outchar);

    fflush(stdout);
    return(0);
}


static
CursorRight(n)
int n;
{
    /** move the cursor 'n' characters to the right (nondestructive) **/
    /** Caller must check that _right is not null before calling **/

    _col = (_col+n < COLUMNS? _col + n: COLUMNS);    /* right 'n' chars... */

    while (n-- > 0)
        tputs(_right, 1, outchar);

    fflush(stdout);
    return(0);
}

static
moveabsolute(col, row)
{

    char *stuff, *tgoto();

    stuff = tgoto(_moveto, col, row);
    tputs(stuff, 1, outchar);
    fflush(stdout);
}

MoveCursor(col, row)
int col, row;
{
int wrap=0; /* Flag to keep track of weither or not we WRAP at edge */
    /** move cursor to the specified row column on the screen.
            0,0 is the top left! **/

if (col < 0)
    {
    col = COLUMNS-1;
    wrap = 1;
    }
if (col >= COLUMNS)
    {
      col = 0;
    wrap = 2;
    }
if (row < 0)
    {
      return SCROLLUP;
    }
if (row > LINES)
    {
    return SCROLLDOWN;
    }

if (!_moveto)
    return -1;

moveabsolute(col, row);
if (wrap == 1) return WRAPLEFT;
if (wrap == 2) return WRAPRIGHT;
return 0;
}

CarriageReturn()
{
    /** move the cursor to the beginning of the current line **/
    Writechar('\r');
}

NewLine()
{
    /** move the cursor to the beginning of the next line **/

    Writechar('\r');
    Writechar('\n');
}

KeypadOn()
{
    if (!_transmit_on)
        return(-1);

    tputs(_transmit_on, 1, outchar);
    fflush(stdout);
    return(0);
}

KeypadOff()
{
    if (!_transmit_off)
        return(-1);

    tputs(_transmit_off, 1, outchar);
    fflush(stdout);
    return(0);
}

TermBegin()
{
    if (!_termbegin)
        return(-1);

    tputs(_termbegin, 1, outchar);
    fflush(stdout);
    return(0);
}

TermEnd()
{
    if (!_termend)
        return(-1);

    tputs(_termend, 1, outchar);
    fflush(stdout);
    return(0);
}

StartBold()
{
    /** start boldface/standout mode **/

    if (!_setbold)
        return(-1);

    tputs(_setbold, 1, outchar);
    fflush(stdout);
    return(0);
}


EndBold()
{
    /** compliment of startbold **/

    if (!_clearbold)
        return(-1);

    tputs(_clearbold, 1, outchar);
    fflush(stdout);
    return(0);
}

#ifndef ELM

StartUnderline()
{
    /** start underline mode **/

    if (!_setunderline)
        return(-1);

    tputs(_setunderline, 1, outchar);
    fflush(stdout);
    return(0);
}


EndUnderline()
{
    /** the compliment of start underline mode **/

    if (!_clearunderline)
        return(-1);

    tputs(_clearunderline, 1, outchar);
    fflush(stdout);
    return(0);
}


StartHalfbright()
{
    /** start half intensity mode **/

    if (!_sethalfbright)
        return(-1);

    tputs(_sethalfbright, 1, outchar);
    fflush(stdout);
    return(0);
}

EndHalfbright()
{
    /** compliment of starthalfbright **/

    if (!_clearhalfbright)
        return(-1);

    tputs(_clearhalfbright, 1, outchar);
    fflush(stdout);
    return(0);
}

StartInverse()
{
    /** set inverse video mode **/

    if (!_setinverse)
        return(-1);

    tputs(_setinverse, 1, outchar);
    fflush(stdout);
    return(0);
}


EndInverse()
{
    /** compliment of startinverse **/

    if (!_clearinverse)
        return(-1);

    tputs(_clearinverse, 1, outchar);
    fflush(stdout);
    return(0);
}

int
HasMemlock()
{
    /** returns TRUE iff memory locking is available (a terminal
        feature that allows a specified portion of the screen to
        be "locked" & not cleared/scrolled... **/

    return ( _set_memlock && _clear_memlock );
}

static int _old_LINES;

int
StartMemlock()
{

    if (! _set_memlock)
      return(-1);

    if (! _memory_locked) {

      tputs(_set_memlock, 1, outchar);
      fflush(stdout);
      _memory_locked = TRUE;
    }

    return(0);
}

int
EndMemlock()
{
    /** Clear the locked memory condition...  **/

    if (! _set_memlock)
      return(-1);

    if(_memory_locked) {
      tputs(_clear_memlock, 1, outchar);
      fflush(stdout);
      _memory_locked = FALSE;
    }
    return(0);
}

#endif /* ndef ELM */

int Writechar(ch,locx, locy, vlocy)
register int ch;
int *locx;
int *locy;
int *vlocy;
{
extern int MAXLINE;
extern int MODE;

    /** write a character to the current screen location. **/

      /* if we only have one column left, simulate automargins if
       * the terminal doesn't have them */

putchar(ch);
if(*locx == COLUMNS - 1)
    {
    *locx=0;
    *locy+=1;
    *vlocy+=1;
         if (!_automargin || _eatnewlineglitch)
        {
              putchar('\r');
             putchar('\n');
            }
    if (*vlocy > MAXLINE-1)
        {
        *locy-=1;
        *vlocy-=1;
        MoveCursor(*locx,*locy);
        return 1;
        }
    if (*locy > LINES -1)
        {
          return SCROLLDOWN;
        }
    }
else
    {
    *locx+=1;
    }
return 1;
}

/*VARARGS2*/

Write_to_screen(line, argcount, arg1, arg2, arg3)
char *line;
int   argcount;
char *arg1, *arg2, *arg3;
{
    /** This routine writes to the screen at the current location.
          when done, it increments lines & columns accordingly by
        looking for "\n" sequences... **/

    switch (argcount) {
    case 0 :
        PutLine0(_line, _col, line);
        break;
    case 1 :
        PutLine1(_line, _col, line, arg1);
        break;
    case 2 :
        PutLine2(_line, _col, line, arg1, arg2);
        break;
    case 3 :
        PutLine3(_line, _col, line, arg1, arg2, arg3);
        break;
    }
}

PutLine0(x, y, line)
int x,y;
register char *line;
{
    /** Write a zero argument line at location x,y **/

    MoveCursor(x,y);
    while(*line)
      Writechar(*line++);
    fflush(stdout);
}

/*VARARGS2*/
PutLine1(x,y, line, arg1)
int x,y;
char *line;
char *arg1;
{
    /** write line at location x,y - one argument... **/

    char buffer[VERY_LONG_STRING];

    sprintf(buffer, line, arg1);

    PutLine0(x, y, buffer);
        fflush(stdout);
}

/*VARARGS2*/
PutLine2(x,y, line, arg1, arg2)
int x,y;
char *line;
char *arg1, *arg2;
{
    /** write line at location x,y - one argument... **/

    char buffer[VERY_LONG_STRING];

    sprintf(buffer, line, arg1, arg2);

    PutLine0(x, y, buffer);
        fflush(stdout);
}

/*VARARGS2*/
PutLine3(x,y, line, arg1, arg2, arg3)
int x,y;
char *line;
char *arg1, *arg2, *arg3;
{
    /** write line at location x,y - one argument... **/

    char buffer[VERY_LONG_STRING];

    sprintf(buffer, line, arg1, arg2, arg3);

    PutLine0(x, y, buffer);
        fflush(stdout);
}

CleartoEOLN()
{
    /** clear to end of line **/

    if (!_cleartoeoln)
        return(-1);

    tputs(_cleartoeoln, 1, outchar);
    fflush(stdout);  /* clear the output buffer */
    return(0);
}

CleartoEOS()
{
    /** clear to end of screen **/

    if (!_cleartoeos)
        return(-1);

    tputs(_cleartoeos, 1, outchar);
    fflush(stdout);  /* clear the output buffer */
    return(0);
}


RawState()
{
    /** returns either 1 or 0, for ON or OFF **/

    return( _inraw );
}

Raw(state)
int state;
{
    /** state is either ON or OFF, as indicated by call **/

    if (state == OFF && _inraw) {
      (void) ttsetattr(TTYIN,&_original_tty);
      _inraw = 0;
    TermEnd();
    KeypadOff();
    }
    else if (state == ON && ! _inraw) {

      (void) ttgetattr(TTYIN, &_original_tty);
      (void) ttgetattr(TTYIN, &_raw_tty);    /** again! **/

#if !defined(TERMIO) && !defined(TERMIOS)
      _raw_tty.sg_flags &= ~(ECHO);    /* echo off */

      _raw_tty.sg_flags |= RAW;    /* raw on    */
#else

#ifdef u3b2
      _raw_tty.c_lflag &= ~(ICANON | ISIG | ECHO);
#else
      _raw_tty.c_lflag &= ~(ICANON | ISIG | ECHO | IEXTEN);
#endif

      _raw_tty.c_cc[VMIN] = '\01';    /* minimum # of chars to queue    */
      _raw_tty.c_cc[VTIME] = '\0';    /* minimum time to wait for input */

#endif
      (void) ttsetattr(TTYIN, &_raw_tty);
      _inraw = 1;

    TermBegin();
    KeypadOn();
    }
}

int
ReadCh()
{
    /** read a character with Raw mode set! **/

    register int result;
    char ch;
    result = read(0, &ch, 1);
        return((result <= 0 ) ? EOF : ch);
}

outchar(c)
char c;
{
    /** output the given character.  From tputs... **/
    /** Note: this CANNOT be a macro!              **/

    putc(c, stdout);
}

int
chloc(string, ch)
char *string, ch;
{
        /** returns the index of ch in string, or -1 if not in string **/
        register int i, len;

        for (i=0, len = strlen(string); i<len; i++)
          if (string[i] == ch) return(i);
        return(-1);
}


InsertLine()
{
    if (!_insert_line)
        {
                putchar(27);
        putchar('[');
        putchar('L');
        fflush(stdout);
        return(-1);
        }

    tputs(_insert_line, 1, outchar);
    fflush(stdout);
    return(0);
}

DeleteLine()
{

    if (!_delete_line)
        {
        putchar(27);
        putchar('[');
        putchar('M');
        fflush(stdout);
        return(-1);
        }

    tputs(_delete_line, 1, outchar);
    fflush(stdout);
    return(0);
}

